from mygrad.tensor_base import Tensor
import numpy as np
from numpy.testing import assert_allclose, assert_array_equal

from hypothesis import given, settings
import hypothesis.strategies as st
import hypothesis.extra.numpy as hnp

from ..utils.numerical_gradient import numerical_gradient_full
from ..custom_strategies import basic_index, adv_integer_index, broadcastable_shape

from mygrad.tensor_core_ops.indexing import _arr, _is_bool_array_index, _is_int_array_index


# test utilties used by setitem
def test_arr_util():
    assert_array_equal(_arr(2, 2), np.arange(4).reshape(2, 2))
    assert_array_equal(_arr(4, 3), np.arange(12).reshape(4, 3))


def test_int_array_test():
    assert _is_int_array_index((0, 0)) is False
    assert _is_int_array_index((np.array([True]), )) is False
    assert _is_int_array_index((np.array([True]), [1])) is True
    assert _is_int_array_index((np.array([1]), [1])) is True
    assert _is_int_array_index((np.array([True]), 1)) is False
    assert _is_int_array_index((np.array([True]), slice(None))) is False


def test_bool_array_test():
    assert _is_bool_array_index((0, 0)) is False
    assert _is_bool_array_index((np.array([True]),)) is True
    assert _is_bool_array_index((np.array([True]), np.array([False]))) is False
    assert _is_bool_array_index((np.array([1]), [1])) is False
    assert _is_bool_array_index((np.array([True]), 1)) is False
    assert _is_bool_array_index((np.array([True]), slice(None))) is False


def setitem(x, y, index):
    x_copy = np.copy(x)
    x_copy[index] = y
    return x_copy


def test_setitem_multiple_input():
    """
    Ensures proper backprop through computational graph
    in which variable that is set on serves as multiple
    inputs to a single operation.

    Ensures that null-gradient and clear-graph works properly.
    """
    from mygrad import add_sequence
    x = Tensor([1.])
    y = x + 0

    assert_array_equal(y.data, np.array([1.]))

    o = add_sequence(y, y, y)
    y[0] = 4

    assert_array_equal(y.data, np.array([4.]))

    f = o * y  # 3 * 4
    f.backward()

    assert_array_equal(o.data, np.array([3.]))
    assert_array_equal(f.data, np.array([12.]))

    assert_array_equal(x.grad, np.array([12.]))
    assert_array_equal(o.grad, np.array([4.]))
    assert_array_equal(y.grad, np.array([3.]))

    f.null_gradients()
    assert x.grad is None and not x._ops and not x._accum_ops
    assert y.grad is None and not y._ops and not y._accum_ops
    assert o.grad is None and not o._ops and not o._accum_ops
    assert f.grad is None and not f._ops and not f._accum_ops
    

@given(x_constant=st.booleans(),
       y_constant=st.booleans(),
       data=st.data())
def test_setitem_sanity_check(x_constant, y_constant, data):
    """ Ensure proper setitem behavior for all combinations of constant/variable Tensors"""
    x = Tensor([1., 2., 3., 4.], constant=x_constant)
    w = 4 * x

    as_tensor = data.draw(st.booleans()) if y_constant else True
    y = Tensor([1., 0.], constant=y_constant) if as_tensor else np.array([1., 0.])

    w[::2] = np.array([-1., -2.]) * y
    assert_allclose(np.array((-1., 8., 0., 16.)), w.data)
    w.sum().backward()

    assert isinstance(w, Tensor)
    assert_allclose(w.data, np.array([-1., 8., 0., 16.]))
    assert w.constant is (x.constant and (not as_tensor or y.constant))

    if x.constant:
        assert x.grad is None
    else:
        assert_allclose(x.grad, np.array([0., 4., 0., 4.]))

    if as_tensor:
        if y.constant:
            assert y.grad is None
        else:
            assert_allclose(y.grad, np.array([-1., -2.]))

    w.null_gradients()
    assert x.grad is None, "null_gradients failed"

    if as_tensor:
        assert y.grad is None, "null_gradients failed"


def test_setitem_sanity_check2():
    x = Tensor([1., 2., 3., 4.])
    y = Tensor([-1., -2., -3., -4.])

    z = x * y
    y[:] = 0

    z.backward()
    assert_allclose(np.array([-1., -2., -3., -4.]), x.grad)
    assert_allclose(np.array([0., 0., 0., 0.]), y.data)
    assert y.grad is None


def test_no_mutate():
    """ Ensure setitem doesn't mutate variable non-constant tensor"""
    x = Tensor([1., 2.])
    y = Tensor([3., 4.])
    x + y
    y[:] = 0
    y_old = x._ops.pop().variables[-1]  # version of y that participated in x + y
    assert_allclose(np.array([3., 4.]), y_old.data)
    assert_allclose(np.array([0., 0.]), y.data)


@settings(deadline=None)
@given(x=hnp.arrays(shape=hnp.array_shapes(max_side=4, max_dims=5),
                    dtype=float,
                    elements=st.floats(-10., 10.)),
       data=st.data())
def test_setitem_basic_index(x, data):
    """ index conforms strictly to basic indexing """
    index = data.draw(basic_index(x.shape), label="index")
    o = np.asarray(x[index])
    y = data.draw(hnp.arrays(shape=broadcastable_shape(o.shape, max_dim=o.ndim),
                             dtype=float,
                             elements=st.floats(-10., 10.)), label="y")

    grad = data.draw(hnp.arrays(shape=x.shape,
                                dtype=float,
                                elements=st.floats(1, 10),
                                unique=True),
                     label="grad")

    x0 = np.copy(x)
    y0 = np.copy(y)

    x_arr = Tensor(np.copy(x))
    y_arr = Tensor(np.copy(y))
    x1_arr = x_arr[:]
    x1_arr[index] = y_arr
    (x1_arr * grad).sum().backward()

    x0[index] = y0
    assert_allclose(x1_arr.data, x0)
    assert_allclose(y_arr.data, y0)

    dx, dy = numerical_gradient_full(setitem, x, y, back_grad=grad,
                                     as_decimal=True, kwargs=dict(index=index))

    assert_allclose(x_arr.grad, dx)
    assert_allclose(y_arr.grad, dy)


@settings(deadline=None)
@given(x=hnp.arrays(shape=hnp.array_shapes(max_side=4, max_dims=5),
                    dtype=float,
                    elements=st.floats(-10., 10.)),
       data=st.data())
def test_setitem_adv_int_index(x, data):
    """ index consists of a tuple of integer-valued arrays """
    index = data.draw(adv_integer_index(x.shape), label="index")
    o = np.asarray(x[index])
    y = data.draw(hnp.arrays(shape=broadcastable_shape(o.shape, max_dim=o.ndim),
                             dtype=float,
                             elements=st.floats(-10., 10.)), label="y")

    grad = data.draw(hnp.arrays(shape=x.shape,
                                dtype=float,
                                elements=st.floats(1, 10),
                                unique=True),
                     label="grad")

    x0 = np.copy(x)
    y0 = np.copy(y)

    x_arr = Tensor(np.copy(x))
    y_arr = Tensor(np.copy(y))
    x1_arr = x_arr[:]
    x1_arr[index] = y_arr
    (x1_arr * grad).sum().backward()

    x0[index] = y0
    assert_allclose(x1_arr.data, x0)
    assert_allclose(y_arr.data, y0)

    dx, dy = numerical_gradient_full(setitem, x, y, back_grad=grad,
                                     as_decimal=True, kwargs=dict(index=index))

    assert_allclose(x_arr.grad, dx)
    assert_allclose(y_arr.grad, dy)


@settings(deadline=None)
@given(x=hnp.arrays(shape=hnp.array_shapes(max_side=4, max_dims=5),
                    dtype=float,
                    elements=st.floats(-10., 10.)),
       data=st.data())
def test_setitem_adv_bool_index(x, data):
    """ index consists of a single boolean-valued array """
    index = data.draw(hnp.arrays(shape=x.shape, dtype=bool), label="index")
    o = np.asarray(x[index])
    y = data.draw(hnp.arrays(shape=broadcastable_shape(o.shape, max_dim=o.ndim),
                             dtype=float,
                             elements=st.floats(-10., 10.)), label="y")

    grad = data.draw(hnp.arrays(shape=x.shape,
                                dtype=float,
                                elements=st.floats(1, 10),
                                unique=True),
                     label="grad")

    x0 = np.copy(x)
    y0 = np.copy(y)

    x_arr = Tensor(np.copy(x))
    y_arr = Tensor(np.copy(y))
    x1_arr = x_arr[:]
    x1_arr[index] = y_arr
    (x1_arr * grad).sum().backward()

    x0[index] = y0
    assert_allclose(x1_arr.data, x0)
    assert_allclose(y_arr.data, y0)

    dx, dy = numerical_gradient_full(setitem, x, y, back_grad=grad,
                                     as_decimal=True, kwargs=dict(index=index))

    assert_allclose(x_arr.grad, dx)
    assert_allclose(y_arr.grad, dy)


@settings(deadline=None)
@given(x=hnp.arrays(shape=(4, 3),
                    dtype=float,
                    elements=st.floats(-10., 10.)),
       data=st.data())
def test_setitem_broadcast_index(x, data):
    """ index is two broadcast-compatible integer arrays"""
    # test broadcast-compatible int-arrays
    rows = np.array([0, 3], dtype=np.intp)
    columns = np.array([0, 2], dtype=np.intp)
    index = np.ix_(rows, columns)
    o = np.asarray(x[index])
    y = data.draw(hnp.arrays(shape=broadcastable_shape(o.shape, max_dim=o.ndim),
                             dtype=float,
                             elements=st.floats(-10., 10.)), label="y")

    grad = data.draw(hnp.arrays(shape=x.shape,
                                dtype=float,
                                elements=st.floats(1, 10),
                                unique=True),
                     label="grad")

    x0 = np.copy(x)
    y0 = np.copy(y)

    x_arr = Tensor(np.copy(x))
    y_arr = Tensor(np.copy(y))
    x1_arr = x_arr[:]
    x1_arr[index] = y_arr
    (x1_arr * grad).sum().backward()

    x0[index] = y0
    assert_allclose(x1_arr.data, x0)
    assert_allclose(y_arr.data, y0)

    dx, dy = numerical_gradient_full(setitem, x, y, back_grad=grad,
                                     as_decimal=True, kwargs=dict(index=index))

    assert_allclose(x_arr.grad, dx)
    assert_allclose(y_arr.grad, dy)


@settings(deadline=None)
@given(x=hnp.arrays(shape=(4, 3),
                    dtype=float,
                    elements=st.floats(-10., 10.)),
       data=st.data())
def test_setitem_mixed_index(x, data):
    """ index is mixes basic and advanced int-array indexing"""
    index = (slice(1, 2), [1, 2])
    o = np.asarray(x[index])
    y = data.draw(hnp.arrays(shape=broadcastable_shape(o.shape, max_dim=o.ndim),
                             dtype=float,
                             elements=st.floats(-10., 10.)), label="y")

    grad = data.draw(hnp.arrays(shape=x.shape,
                                dtype=float,
                                elements=st.floats(1, 10),
                                unique=True),
                     label="grad")

    x0 = np.copy(x)
    y0 = np.copy(y)

    x_arr = Tensor(np.copy(x))
    y_arr = Tensor(np.copy(y))
    x1_arr = x_arr[:]
    x1_arr[index] = y_arr
    (x1_arr * grad).sum().backward()

    x0[index] = y0
    assert_allclose(x1_arr.data, x0)
    assert_allclose(y_arr.data, y0)

    dx, dy = numerical_gradient_full(setitem, x, y, back_grad=grad,
                                     as_decimal=True, kwargs=dict(index=index))

    assert_allclose(x_arr.grad, dx)
    assert_allclose(y_arr.grad, dy)


@settings(deadline=None)
@given(x=hnp.arrays(shape=(4, 3),
                    dtype=float,
                    elements=st.floats(-10., 10.)),
       data=st.data())
def test_setitem_broadcast_bool_index(x, data):
    """ index mixes boolean and int-array indexing"""
    rows = np.array([False,  True, False,  True])
    columns = np.array([0, 2], dtype=np.intp)
    index = np.ix_(rows, columns)
    o = np.asarray(x[index])
    y = data.draw(hnp.arrays(shape=broadcastable_shape(o.shape, max_dim=o.ndim),
                             dtype=float,
                             elements=st.floats(-10., 10.)), label="y")

    grad = data.draw(hnp.arrays(shape=x.shape,
                                dtype=float,
                                elements=st.floats(1, 10),
                                unique=True),
                     label="grad")

    x0 = np.copy(x)
    y0 = np.copy(y)

    x_arr = Tensor(np.copy(x))
    y_arr = Tensor(np.copy(y))
    x1_arr = x_arr[:]
    x1_arr[index] = y_arr
    (x1_arr * grad).sum().backward()

    x0[index] = y0
    assert_allclose(x1_arr.data, x0)
    assert_allclose(y_arr.data, y0)

    dx, dy = numerical_gradient_full(setitem, x, y, back_grad=grad,
                                     as_decimal=True, kwargs=dict(index=index))

    assert_allclose(x_arr.grad, dx)
    assert_allclose(y_arr.grad, dy)


@settings(deadline=None)
@given(x=hnp.arrays(shape=(4, 3),
                    dtype=float,
                    elements=st.floats(-10., 10.)),
       data=st.data())
def test_setitem_bool_basic_index(x, data):
    """ index mixes boolean and basic indexing"""
    index = (np.array([False,  True, False,  True]), np.newaxis, slice(None))
    o = np.asarray(x[index])
    y = data.draw(hnp.arrays(shape=broadcastable_shape(o.shape, max_dim=o.ndim),
                             dtype=float,
                             elements=st.floats(-10., 10.)), label="y")

    grad = data.draw(hnp.arrays(shape=x.shape,
                                dtype=float,
                                elements=st.floats(1, 10),
                                unique=True),
                     label="grad")

    x0 = np.copy(x)
    y0 = np.copy(y)

    x_arr = Tensor(np.copy(x))
    y_arr = Tensor(np.copy(y))
    x1_arr = x_arr[:]
    x1_arr[index] = y_arr
    (x1_arr * grad).sum().backward()

    x0[index] = y0
    assert_allclose(x1_arr.data, x0)
    assert_allclose(y_arr.data, y0)

    dx, dy = numerical_gradient_full(setitem, x, y, back_grad=grad,
                                     as_decimal=True, kwargs=dict(index=index))

    assert_allclose(x_arr.grad, dx)
    assert_allclose(y_arr.grad, dy)


@settings(deadline=None)
@given(x=hnp.arrays(shape=(3, 3),
                    dtype=float,
                    elements=st.floats(-10., 10.)),
       data=st.data())
def test_setitem_bool_axes_index(x, data):
    """ index consists of boolean arrays specified for each axis """
    index = data.draw(st.tuples(hnp.arrays(shape=(3,), dtype=bool), hnp.arrays(shape=(3,), dtype=bool)))
    try:
        o = np.asarray(x[index])
    except IndexError:
        return None
    y = data.draw(hnp.arrays(shape=broadcastable_shape(o.shape, max_dim=o.ndim),
                             dtype=float,
                             elements=st.floats(-10., 10.)), label="y")

    grad = data.draw(hnp.arrays(shape=x.shape,
                                dtype=float,
                                elements=st.floats(1, 10),
                                unique=True),
                     label="grad")

    x0 = np.copy(x)
    y0 = np.copy(y)

    x_arr = Tensor(np.copy(x))
    y_arr = Tensor(np.copy(y))
    x1_arr = x_arr[:]
    x1_arr[index] = y_arr
    (x1_arr * grad).sum().backward()

    x0[index] = y0
    assert_allclose(x1_arr.data, x0)
    assert_allclose(y_arr.data, y0)

    dx, dy = numerical_gradient_full(setitem, x, y, back_grad=grad,
                                     as_decimal=True, kwargs=dict(index=index))

    assert_allclose(x_arr.grad, dx)
    assert_allclose(y_arr.grad, dy)



""" Custom hypothesis search strategies """
import hypothesis.strategies as st
import hypothesis.extra.numpy as hnp
import numpy as np

__all__ = ["adv_integer_index",
           "broadcastable_shape",
           "choices",
           "valid_axes",
           "basic_index"]


def choices(seq, size, replace=True):
    """Randomly choose elements from `seq`, producing a tuple of length `size`."""
    if size > len(seq) and not replace:
        raise ValueError("`size` must not exceed the length of `seq` when `replace` is `False`")
    if size > len(seq) and not seq:
        raise ValueError("`size` must be 0, given an empty `seq`")
    inds = list(range(len(seq)))
    if replace:
        strat = st.tuples(*[st.sampled_from(inds)]*size)
    else:
        strat = st.permutations(inds)
    return strat.map(lambda x: tuple(seq[i] for i in x[:size]))


@st.composite
def _rand_neg_axis(draw, axes, ndim):
    """Randomly replace axis values with negative counterpart: axis - ndim"""
    size = draw(st.integers(0, len(axes)))
    neg_inds = draw(choices(range(len(axes)), size, replace=False))
    axes = st.just(axes)
    return draw(axes.map(lambda x: tuple(i - ndim if n in neg_inds else i for n, i in enumerate(x))))


@st.composite
def valid_axes(draw, ndim, pos_only=False, single_axis_only=False, permit_none=True):
    """ Hypothesis search strategy: Given array dimensionality, generate valid
        `axis` arguments (including `None`).


        Parameters
        ----------
        ndim : int
            The dimensionality of the array.

        Returns
        -------
        hypothesis.searchstrategy.SearchStrategy
         -> [Union[NoneType, Tuple[int...]]

        Examples
        --------
        >>> valid_axes(4).example()
        (0, 1)
        """
    if isinstance(ndim, (tuple, list)):
        ndim = len(ndim)

    if 0 > ndim:
        raise ValueError("`ndim` must be an integer 0 or greater.")
    if single_axis_only:
        num_axes = draw(st.integers(min_value=0, max_value=1))
    else:
        num_axes = draw(st.integers(min_value=0, max_value=ndim))

    axes = draw(choices(range(ndim), num_axes, replace=False))

    if not pos_only:
        axes = draw(_rand_neg_axis(axes, ndim))

    if single_axis_only and axes:
        axes = axes[0]

    if permit_none:
        return draw(st.none()) if axes == () else axes
    else:
        return axes if axes else tuple(range(ndim))


@st.composite
def broadcastable_shape(draw, shape, min_dim=0, max_dim=5):
    """ Hypothesis search strategy: given an array shape, generate a
        broadcast-compatible shape, specifying the minimum/maximum permissable
        number of dimensions in the resulting shape (both values are inclusive).

        `draw` is a parameter reserved by hypothesis, and should not be specified
        by the user.

        Parameters
        ----------
        shape : Tuple[int, ...]
        min_dim : int
        max_dim : int

        Returns
        -------
        hypothesis.searchstrategy.SearchStrategy
            -> Tuple[int, ...]

        Examples
        --------
        >>> for i in range(5):
        ...    print(broadcastable_shape(shape=(2, 3)).example())
        (1, 3)
        ()
        (2, 3)
        (5, 2, 3)
        (8, 5, 1, 3)
        (3, )
        """
    ndim = draw(st.integers(min_dim, max_dim))
    n_aligned = min(len(shape), ndim)
    n_leading = ndim - n_aligned
    aligned_dims = draw(st.tuples(*(st.sampled_from([1, size]) for size in shape[::-1][:n_aligned])))[::-1]
    leading_dims = draw(st.tuples(*(st.integers(1, 5) for i in range(n_leading))))
    return leading_dims + aligned_dims


def integer_index(size):
    """ Generate a valid integer-index for an axis of a given size,
        either a positive or negative value.

        Parameters
        ----------
        size : int
            Size of the axis for which the index is drawn

        Returns
        -------
        hypothesis.searchstrategy.SearchStrategy
            -> int"""
    return st.sampled_from(list(range(size)) + list(range(-1, -size, -1)))


@st.composite
def slice_index(draw, size):
    """ Hypothesis search strategy: Generate a valid slice-index
        for an axis of a given size. Slices are chosen such that
        most slices will not be empty.

        `draw` is a parameter reserved by hypothesis, and should not be specified
        by the user.

        Parameters
        ----------
        size : int
            Size of the axis for which the index is drawn

        Returns
        -------
        hypothesis.searchstrategy.SearchStrategy
            -> slice"""
    step = draw(st.sampled_from(list(range(1, max(2, size // 2))) + [-1]))
    start = draw(st.sampled_from(range(size + 1)))
    if step > 0:
        stop = draw(st.sampled_from(range(start, size)))
    else:
        stop = draw(st.sampled_from(range(0, start)))

    return slice(start, stop, step)


@st.composite
def basic_index(draw, shape, max_dim=5):
    """ Hypothesis search strategy: given an array shape, generate a
        a valid index for specifying an element/subarray of that array,
        using basic indexing.

        `draw` is a parameter reserved by hypothesis, and should not be specified
        by the user.

        Parameters
        ----------
        shape : Tuple[int, ...]
            The shape of the array whose indices are being generated

        max_dim : int
            The max dimensionality permitted, given the addition of new-axes.

        Returns
        -------
        hypothesis.searchstrategy.SearchStrategy
            -> Tuple[int, ...]
        """
    min_dim = len(shape)
    max_dim = max(min_dim, max_dim)

    index = list(draw(st.tuples(*(st.one_of(integer_index(size), slice_index(size)) for size in shape))))
    if min_dim != max_dim:
        num_newaxis = draw(st.integers(min_dim, max_dim)) - len(index)
        if num_newaxis:
            for ind in draw(choices(range(min_dim + 1), num_newaxis)):
                index.insert(ind, np.newaxis)
    return tuple(index)



@st.composite
def adv_integer_index(draw, shape, max_dim=3):
    """ Hypothesis search strategy: given an array shape, generate a
        a valid index for specifying an element/subarray of that array,
        using advanced indexing with integer-valued arrays.

        `draw` is a parameter reserved by hypothesis, and should not be specified
        by the user.

        Parameters
        ----------
        shape : Tuple[int, ...]
            The shape of the array whose indices are being generated

        max_dim : int
            The max dimensionality permitted for the index-arrays.

        Returns
        -------
        hypothesis.searchstrategy.SearchStrategy
            -> Tuple[numpy.ndarray, ...]
        """
    index_shape = draw(hnp.array_shapes(max_dims=max_dim))
    index = draw(st.tuples(*(hnp.arrays(dtype=int,
                                        shape=index_shape, elements=st.integers(0, size - 1))
                             for size in shape)))

    return index


from tests.wrappers.uber import fwdprop_test_factory, backprop_test_factory
from mygrad import abs, absolute
import numpy as np


def _is_non_zero(x):
    return np.all(np.abs(x.data) > 1e-8)


@fwdprop_test_factory(mygrad_func=abs, true_func=np.abs, num_arrays=1)
def test_abs_fwd():
    pass


@backprop_test_factory(mygrad_func=abs, true_func=np.abs, num_arrays=1,
                       index_to_bnds={0: (-10, 10)}, assumptions=_is_non_zero,
                       atol=1e-5)
def test_abs_backward():
    pass


@fwdprop_test_factory(mygrad_func=absolute, true_func=np.absolute, num_arrays=1)
def test_absolute_fwd():
    pass


@backprop_test_factory(mygrad_func=absolute, true_func=np.absolute, num_arrays=1,
                       index_to_bnds={0: (-100, 100)}, index_to_no_go={0: (0,)},
                       atol=1e-5, assumptions=_is_non_zero)
def test_absolute_backward():
    pass



from tests.wrappers.uber import fwdprop_test_factory, backprop_test_factory
from mygrad import sinh, cosh, tanh, csch, sech, coth

import numpy as np


def _is_nonzero(x):
    return np.all(np.abs(x.data) > 1e-8)


@fwdprop_test_factory(mygrad_func=sinh, true_func=np.sinh,
                      index_to_bnds={0: (-10, 10)}, num_arrays=1)
def test_sinh_fwd():
    pass


@backprop_test_factory(mygrad_func=sinh, true_func=np.sinh,
                       index_to_bnds={0: (-10, 10)}, num_arrays=1, as_decimal=False)
def test_sinh_backward():
    pass


@fwdprop_test_factory(mygrad_func=cosh, true_func=np.cosh, index_to_bnds={0: (-10, 10)}, num_arrays=1)
def test_cosh_fwd():
    pass


@backprop_test_factory(mygrad_func=cosh, true_func=np.cosh,
                       index_to_bnds={0: (-10, 10)}, atol=1e-5, num_arrays=1, as_decimal=False)
def test_cosh_backward():
    pass


@fwdprop_test_factory(mygrad_func=tanh, true_func=np.tanh, index_to_bnds={0: (-10, 10)}, num_arrays=1)
def test_tanh_fwd():
    pass


@backprop_test_factory(mygrad_func=tanh, true_func=np.tanh,
                       index_to_bnds={0: (-10, 10)}, atol=1e-5, num_arrays=1, as_decimal=False)
def test_tanh_backward():
    pass


@fwdprop_test_factory(mygrad_func=csch, true_func=lambda x: 1 / np.sinh(x), index_to_bnds={0: (.001, 10)}, num_arrays=1)
def test_csch_fwd():
    pass


@backprop_test_factory(mygrad_func=csch, true_func=lambda x: 1 / np.sinh(x),
                       index_to_bnds={0: (.001, 10)}, num_arrays=1, as_decimal=False)
def test_csch_backward():
    pass


@fwdprop_test_factory(mygrad_func=sech, true_func=lambda x: 1 / np.cosh(x),
                      index_to_bnds={0: (-10, 10)}, num_arrays=1)
def test_sech_fwd():
    pass


@backprop_test_factory(mygrad_func=sech, true_func=lambda x: 1 / np.cosh(x),
                       index_to_bnds={0: (.001, 10)}, atol=1e-5, num_arrays=1, as_decimal=False)
def test_sech_backward():
    pass


@fwdprop_test_factory(mygrad_func=coth, true_func=lambda x: 1 / np.tanh(x),
                      index_to_bnds={0: (-10, 10)}, assumptions=_is_nonzero, num_arrays=1)
def test_coth_fwd():
    pass


@backprop_test_factory(mygrad_func=coth, true_func=lambda x: 1 / np.tanh(x),
                       index_to_bnds={0: (.001, 10)},
                       atol=1e-5, num_arrays=1, as_decimal=False)
def test_coth_backward():
    pass



from mygrad.nnet.layers.utils import sliding_window_view

import numpy as np
from numpy.testing import assert_allclose

import hypothesis.strategies as st
from hypothesis import given
import hypothesis.extra.numpy as hnp


dtype_strat_numpy = st.sampled_from((np.int8, np.int16, np.int32, np.int64,
                                     np.float16, np.float32, np.float64))


@given(data=st.data(),
       x=hnp.arrays(dtype=dtype_strat_numpy, shape=hnp.array_shapes(max_dims=5, min_dims=1, max_side=20)))
def test_sliding_window(data, x):
    """ Test variations of window-shape, step, and dilation for sliding window
        view of N-dimensional array."""

    win_dim = data.draw(st.integers(1, x.ndim), label="win_dim")
    win_shape = data.draw(st.tuples(*(st.integers(1, s) for s in x.shape[-win_dim:])),
                          label="win_shape")
    step = data.draw(st.tuples(*(st.integers(1, s) for s in x.shape[-win_dim:])),
                     label="step")

    max_dilation = np.array(x.shape[-win_dim:]) // win_shape
    dilation = data.draw(st.one_of(st.none(), st.tuples(*(st.integers(1, s) for s in max_dilation))),
                         label="dilation")
    y = sliding_window_view(x, window_shape=win_shape, step=step, dilation=dilation)

    if dilation is None:
        dilation = np.ones((len(win_shape),), dtype=int)

    for ind in np.ndindex(*y.shape[:win_dim]):
        slices = tuple(slice(i*s, i*s + w*d, d) for i,w,s,d in zip(ind, win_shape, step, dilation))
        assert_allclose(actual=y[tuple([*ind])], desired=x[(..., *slices)])


@given(dtype=dtype_strat_numpy)
def test_memory_details(dtype):
    """ Ensure that:
          - function handles non C-contiguous layouts correctly
          - output is view of input
          - output is not writeable"""
    x = np.arange(20).reshape(2, 10).astype(dtype)
    x = np.asfortranarray(x)
    y = sliding_window_view(x, (5,), 5)
    soln = np.array([[[ 0,  1,  2,  3,  4],
                      [10, 11, 12, 13, 14]],
                     [[ 5,  6,  7,  8,  9],
                      [15, 16, 17, 18, 19]]])
    assert not y.flags["WRITEABLE"]
    assert_allclose(actual=y, desired=soln)

    x = np.arange(20).reshape(2, 10)
    x = np.ascontiguousarray(x)
    y = sliding_window_view(x, (5,), 5)
    assert not y.flags["WRITEABLE"]
    assert np.shares_memory(x, y)
    assert_allclose(actual=y, desired=soln)




from .conv import conv_nd
from .pooling import max_pool
from .batchnorm import batchnorm


__all__ = ["conv_nd", "max_pool", "batchnorm"]


try:
    from .gru import gru
    from .recurrent import simple_RNN
    __all__ += ['gru', 'simple_RNN']
except ImportError:
    pass





import hypothesis.extra.numpy as hnp
import hypothesis.strategies as st
import numpy as np
from numpy.testing import assert_allclose
from hypothesis import given

from mygrad.tensor_base import Tensor
from mygrad import reshape


def gen_shape(size):
    """ Given an array's size, generate a compatible random shape

        Parameters
        ----------
        size : Integral

        Returns
        -------
        Tuple[int, ...] """

    def _factors(n):
        """ Returns the divisors of n

            >>> _factors(4)
            {1, 2, 4}"""
        from functools import reduce
        return set(reduce(list.__add__,
                          ([i, n // i] for i in range(1, int(n ** 0.5) + 1) if n % i == 0)))

    assert size > 0
    if size == 1:
        return (1,)

    shape = []
    rem = int(size / np.prod(shape))
    while rem > 1:
        if len(shape) > 6:
            shape.append(rem)
            break

        shape.append(np.random.choice(list(_factors(rem))))
        rem = int(size / np.prod(shape))

    return tuple(int(i) for i in shape)


@given(a=hnp.arrays(shape=hnp.array_shapes(max_side=3, max_dims=5),
                    dtype=float,
                    elements=st.floats(-100, 100))
       )
def test_reshape_method_fwd(a):
    new_shape = gen_shape(a.size)

    x = Tensor(a).reshape(new_shape)
    a = a.reshape(new_shape)

    assert x.shape == a.shape, "Tensor.reshape failed"
    assert_allclose(a, x.data), "Tensor.reshape failed"


@given(a=hnp.arrays(shape=hnp.array_shapes(max_side=3, max_dims=5),
                    dtype=float,
                    elements=st.floats(-100, 100))
       )
def test_reshape_method_backward(a):
    new_shape = gen_shape(a.size)
    grad = np.arange(a.size).reshape(a.shape)

    x = Tensor(a)
    o = x.reshape(new_shape)
    o.backward(grad.reshape(new_shape))

    assert x.grad.shape == grad.shape
    assert_allclose(x.grad, grad)


@given(a=hnp.arrays(shape=hnp.array_shapes(max_side=3, max_dims=5),
                    dtype=float,
                    elements=st.floats(-100, 100))
       )
def test_reshape_fwd(a):
    new_shape = gen_shape(a.size)

    x = Tensor(a)
    x = reshape(x, new_shape, constant=True)
    a = a.reshape(new_shape)

    assert x.shape == a.shape, "Tensor.reshape failed"
    assert_allclose(a, x.data), "Tensor.reshape failed"


@given(a=hnp.arrays(shape=hnp.array_shapes(max_side=3, max_dims=5),
                    dtype=float,
                    elements=st.floats(-100, 100))
       )
def test_reshape_backward(a):
    new_shape = gen_shape(a.size)
    grad = np.arange(a.size).reshape(a.shape)

    x = Tensor(a)
    o = reshape(x, new_shape, constant=False)
    o.backward(grad.reshape(new_shape))

    assert x.grad.shape == grad.shape
    assert_allclose(x.grad, grad)



import hypothesis.strategies as st
from numpy.testing import assert_allclose
from hypothesis import given

from mygrad.tensor_base import Tensor


@given(x=st.floats(min_value=-1E-6, max_value=1E6),
       y=st.floats(min_value=-1E-6, max_value=1E6),
       z=st.floats(min_value=-1E-6, max_value=1E6),
       side_effects=st.booleans())
def test_chainrule_scalar(x, y, z, side_effects):
    x = Tensor(x)
    y = Tensor(y)
    z = Tensor(z)

    f = x*y + z
    g = x + z*f*f

    if side_effects:
        # check side effects
        unused = 2*g - f
        w = 1*f
    else:
        unused = Tensor(0)
        w = Tensor(0)

    g.backward()
    assert_allclose(f.grad, 2 * z.data * f.data)
    assert_allclose(x.grad, 1 + 2 * z.data * f.data * y.data)
    assert_allclose(y.grad, 2 * z.data * f.data * x.data)
    assert_allclose(z.grad, f.data**2 + z.data * 2 * f.data)

    assert w.grad is None



""" Test all binary arithmetic operations, checks for appropriate broadcast behavior"""
from ...wrappers.uber import fwdprop_test_factory, backprop_test_factory

from mygrad import add, subtract, multiply, divide, power, logaddexp
from mygrad import logaddexp2, maximum, minimum
from mygrad import arctan2

import numpy as np


from mygrad.tensor_base import Tensor
from numpy.testing import assert_allclose

from hypothesis import settings


@fwdprop_test_factory(mygrad_func=add, true_func=np.add, num_arrays=2)
def test_add_fwd():
    pass


@backprop_test_factory(mygrad_func=add, true_func=np.add, num_arrays=2,
                       atol=1e-4, rtol=1e-4)
def test_add_bkwd():
    pass


@fwdprop_test_factory(mygrad_func=subtract, true_func=np.subtract, num_arrays=2)
def test_subtract_fwd():
    pass


@backprop_test_factory(mygrad_func=subtract, true_func=np.subtract, num_arrays=2,
                       atol=1e-4, rtol=1e-4)
def test_subtract_bkwd():
    pass


@fwdprop_test_factory(mygrad_func=multiply, true_func=np.multiply, num_arrays=2)
def test_multiply_fwd():
    pass


@backprop_test_factory(mygrad_func=multiply, true_func=np.multiply, atol=1e-4, rtol=1e-4, num_arrays=2)
def test_multiply_bkwd():
    pass


@fwdprop_test_factory(mygrad_func=divide, true_func=np.divide, index_to_bnds={1: (1, 10)}, num_arrays=2)
def test_divide_fwd():
    pass


@backprop_test_factory(mygrad_func=divide, true_func=np.divide, index_to_bnds={1: (1, 10)}, num_arrays=2,
                       atol=1e-4, rtol=1e-4)
def test_divide_bkwd():
    pass


@fwdprop_test_factory(mygrad_func=power, true_func=np.power,
                      index_to_bnds={0: (1, 10), 1: (-3, 3)},
                      num_arrays=2)
def test_power_fwd():
    pass


@settings(deadline=None)
@backprop_test_factory(mygrad_func=power, true_func=np.power,
                       index_to_bnds={0: (1, 10), 1: (-3, 3)},
                       num_arrays=2, atol=1e-4, rtol=1e-4)
def test_power_bkwd():
    pass


@fwdprop_test_factory(mygrad_func=logaddexp, true_func=np.logaddexp, num_arrays=2)
def test_logaddexp_fwd():
    pass


@settings(deadline=None)
@backprop_test_factory(mygrad_func=logaddexp, true_func=np.logaddexp, num_arrays=2,
                       as_decimal=False, atol=1e-4, rtol=1e-4, vary_each_element=True,
                       index_to_bnds={0: (-2, 2), 1: (-2, 2)})
def test_logaddexp_bkwd():
    pass


@fwdprop_test_factory(mygrad_func=logaddexp2, true_func=np.logaddexp2, num_arrays=2)
def test_logaddexp2_fwd():
    pass


@settings(deadline=None)
@backprop_test_factory(mygrad_func=logaddexp2, true_func=np.logaddexp2, num_arrays=2,
                       as_decimal=False, atol=1e-4, rtol=1e-4, vary_each_element=True)
def test_logaddexp2_bkwd():
    pass


@fwdprop_test_factory(mygrad_func=arctan2, true_func=np.arctan2, num_arrays=2, index_to_bnds={1: (1, 10)})
def test_arctan2_fwd():
    pass


@settings(deadline=None)
@backprop_test_factory(mygrad_func=arctan2, true_func=np.arctan2, num_arrays=2,
                       as_decimal=False, atol=1e-4, rtol=1e-4, vary_each_element=True,
                       index_to_bnds={1: (1, 10)})
def test_arctan2_bkwd():
    pass


@fwdprop_test_factory(mygrad_func=maximum, true_func=np.maximum, num_arrays=2)
def test_maximum_fwd():
    pass


def is_not_close(arr0: Tensor, arr1: Tensor) -> bool:
    return not np.any(np.isclose(arr0.data, arr1.data))


@backprop_test_factory(mygrad_func=maximum, true_func=np.maximum, num_arrays=2,
                       assumptions=is_not_close)
def test_maximum_bkwd():
    pass


def test_maximum_bkwd_equal():
    """ regression test for documented behavior of maximum/minimum where
        x == y"""

    x = Tensor([1., 0., 2.])
    y = Tensor([2., 0., 1.])

    o = maximum(x, y)
    o.backward()

    assert_allclose(x.grad, [0., 0., 1])
    assert_allclose(y.grad, [1., 0., 0])
    o.null_gradients()


@fwdprop_test_factory(mygrad_func=minimum, true_func=np.minimum, num_arrays=2)
def test_minimum_fwd(): pass


@backprop_test_factory(mygrad_func=minimum, true_func=np.minimum, num_arrays=2,
                       assumptions=is_not_close)
def test_minimum_bkwd():
    pass


def test_minimum_bkwd_equal():
    """ regression test for documented behavior of minimum/minimum where
        x == y"""

    x = Tensor([1., 0., 2.])
    y = Tensor([2., 0., 1.])

    o = minimum(x, y)
    o.backward()

    assert_allclose(x.grad, [1., 0., 0.])
    assert_allclose(y.grad, [0., 0., 1.])
    o.null_gradients()



from mygrad.operation_base import Operation
from mygrad.tensor_base import Tensor
from numbers import Integral
import numpy as np

try:
    from numba import njit
except ImportError:
    raise ImportError("The package `numba` must be installed in order to access the simple-rnn.")


@njit
def dot(a, b):
    """
    Calculates the dot product between 2 arrays
    of shapes (W,X,Y) and (Y,Z), respectively
    """
    out = np.zeros((a.shape[0], a.shape[1], b.shape[-1]))
    for i in range(len(a)):
        out[i] = np.dot(a[i], b)
    return out


@njit
def _dot_tanh_accum(x, W):
    for n in range(len(x) - 1):
        x[n + 1] += np.dot(x[n], W)
        x[n + 1] = np.tanh(x[n + 1])


@njit
def _rnn_bptt(X, dLt_dst, dst_dft, W, bp_lim, old_dst=None):
    W = W.T
    if bp_lim < len(X) - 1:
        old_dst = np.zeros_like(dLt_dst)

    for i in range(bp_lim):
        if bp_lim < len(X) - 1:
            source_index = slice(2, len(dLt_dst) - i)
            target_index = slice(1, len(dLt_dst) - (i + 1))

            # dL_{n} / ds_{t+1} -> dL_{n} / df_{t+1}  | ( n > t )
            dLn_ft1 = dst_dft[source_index] * (dLt_dst[source_index] - old_dst[source_index])
            old_dst = np.copy(dLt_dst)

        else:  # no backprop truncation
            source_index = slice(len(dLt_dst) - (i + 1), len(dLt_dst) - i)
            target_index = slice(len(dLt_dst) - (i + 2), len(dLt_dst) - (i + 1))
            dLn_ft1 = dst_dft[source_index] * dLt_dst[source_index]

        dLt_dst[target_index] += dot(dLn_ft1, W)


def _backprop(var, grad):
    if not var.constant:
        if var.grad is None:
            var.grad = np.asarray(grad)
        else:
            var.grad += grad

class RecurrentUnit(Operation):
    """ Defines a basic recurrent unit for a RNN.

        This unit operates on a sequence of data {X_j | (0 <= j <= T - 1)}, producing
        a sequence of "hidden descriptors": {S_i | (0 <= i <= T}}, via the trainable parameters
        U and W

                                S_{t} = tanh(U X_{t-1} + W S_{t-1})

        For a language model, S_{t} is traditionally mapped to a prediction of X_t via: softmax(V S_t),
        where V is yet another trainable parameter (not built into the recurrent unit)."""

    scalar_only = True

    def __call__(self, X, U, W, s0=None, bp_lim=None):
        """ Performs a forward pass of sequential data through a simple RNN layer, returning
            the 'hidden-descriptors' arrived at by utilizing the trainable parameters U and V:

                                S_{t} = tanh(U X_{t-1} + W S_{t-1})

            Parameters
            ----------
            X : mygrad.Tensor, shape=(T, N, C)
               The sequential data to be passed forward.

            U : mygrad.Tensor, shape=(D, C)
               The weights used to map sequential data to its hidden-descriptor representation

            W : mygrad.Tensor, shape=(D, D)
                The weights used to map a hidden-descriptor to a hidden-descriptor.

            s0 : Optional[mygrad.Tensor, numpy.ndarray], shape=(N, D)
                The 'seed' hidden descriptors to feed into the RNN. If None, a Tensor
                of zeros of shape (N, D) is created.

            bp_lim : Optional[int]
                The (non-zero) limit of the number of back propagations through time are
                performed


            Returns
            -------
            mygrad.Tensor
                The sequence of 'hidden-descriptors' produced by the forward pass of the RNN."""
        if bp_lim is not None:
            assert isinstance(bp_lim, Integral) and 0 <= bp_lim < len(X)
        self.bp_lim = bp_lim if bp_lim is not None else len(X) - 1

        self.X = X
        self.U = U
        self.W = W
        self.variables = (self.X, self.U, self.W)
        self._hidden_seq = []

        seq = self.X.data
        out = np.zeros((seq.shape[0] + 1, seq.shape[1], self.U.shape[-1]))

        if s0 is not None:
            out[0] = s0.data if isinstance(s0, Tensor) else s0

        out[1:] = dot(seq, self.U.data)
        _dot_tanh_accum(out, self.W.data)

        self._hidden_seq = Tensor(out, _creator=self)

        return self._hidden_seq.data

    def backward(self, grad, *, graph, **kwargs):
        """ Performs back propagation through time (with optional truncation), using the
            following notation:

                f_t = U x_{t-1} + W s_{t-1}
                s_t = tanh(f_t)
        """
        if self.U.constant and self.W.constant and self.X.constant:
            return None

        s = self._hidden_seq

        dst_dft = (1 - s.data ** 2)  # ds_{t} / d_f{t}
        dLt_dst = np.copy(grad)  # dL_{t} / ds_{t}

        _rnn_bptt(self.X.data, dLt_dst, dst_dft, self.W.data, self.bp_lim)

        self._hidden_seq.grad = dLt_dst  # element t: dL_{t} / ds_{t} + ... + dL_{T_lim} / ds_{t}

        dLt_dft = dLt_dst[1:] * dst_dft[1:]  # dL_{t} / df_{t} + ... + dL_{T_lim} / df_{t}

        if not self.U.constant:
            _backprop(self.U, np.einsum("ijk, ijl -> kl", self.X.data, dLt_dft))  # dL_{1} / dU + ... + dL_{T} / dU
        if not self.W.constant:
            _backprop(self.W, np.einsum("ijk, ijl -> kl", s.data[:-1], dLt_dft))  # dL_{1} / dW + ... + dL_{T} / dW
        if not self.X.constant:
            _backprop(self.X, dot(dLt_dft, self.U.data.T))  # dL_{1} / dX + ... + dL_{T} / dX


def simple_RNN(X, U, W, s0=None, bp_lim=None, constant=False):
    """ Performs a forward pass of sequential data through a simple RNN layer, returning
        the 'hidden-descriptors' arrived at by utilizing the trainable parameters U and V::

                            S_{t} = tanh(U X_{t-1} + W S_{t-1})

        Parameters
        ----------
        X : mygrad.Tensor, shape=(T, N, C)
           The sequential data to be passed forward.

        U : mygrad.Tensor, shape=(D, C)
           The weights used to map sequential data to its hidden-descriptor representation

        W : mygrad.Tensor, shape=(D, D)
            The weights used to map a hidden-descriptor to a hidden-descriptor.

        s0 : Optional[mygrad.Tensor, numpy.ndarray], shape=(N, D)
            The 'seed' hidden descriptors to feed into the RNN. If None, a Tensor
            of zeros of shape (N, D) is created.

        bp_lim : Optional[int]
            The (non-zero) limit of the depth of back propagation through time to be
            performed. If `None` back propagation is passed back through the entire sequence.

            E.g. `bp_lim=3` will propagate gradients only up to 3 steps backward through the
            recursive sequence.

        constant : bool, optional (default=False)
            If True, the resulting Tensor is a constant.

        Returns
        -------
        mygrad.Tensor
            The sequence of 'hidden-descriptors' produced by the forward pass of the RNN.

        Notes
        -----
        This simple RNN implements the recursive system of equations:

        .. math::
          S_{t} = tanh(U X_{t-1} + W S_{t-1})

        - :math:`T` : Sequence length
        - :math:`N` : Batch size
        - :math:`C` : Length of single datum
        - :math:`D` : Length of 'hidden' descriptor"""
    s = Tensor._op(RecurrentUnit, X, U, W, op_kwargs=dict(s0=s0, bp_lim=bp_lim),
                   constant=constant)
    s.creator._hidden_seq = s
    return s



import numpy as np

class SimpleOperation:
    def __call__(self, *input_vars):
        self.variables = input_vars
        raise NotImplementedError

    def backward_var(self, grad, index):
        raise NotImplementedError

    def backward(self, grad):
        for index, var in enumerate(self.variables):
            if not var.constant:
                self.backward_var(grad, index)

    def null_gradients(self, clear_graph=True):
        for var in self.variables:
            var.null_gradients(clear_graph=clear_graph)

    def clear_graph(self, terminal_node):
        for var in self.variables:
            var.clear_graph(terminal_node)


class Node:
    __array_priority__ = 15.0

    def __init__(self, x, *, constant=False, _creator=None):
        self.creator = _creator
        self.data = x.data if isinstance(x, Node) else np.asarray(x)
        self.grad = None
        self.constant = constant
        self._ops = []  # Operation instances that utilized self an input tensor

    def __repr__(self):
        return repr(self.data).replace("array", "Node").replace("\n", "\n ")

    @classmethod
    def _op(cls, Op, *input_vars, constant=False):
        tensor_vars = []
        for var in input_vars:
            if not isinstance(var, cls):
                var = cls(var, constant=True)
            tensor_vars.append(var)

        f = Op()
        op_out = f(*tensor_vars)
        is_const = constant or all(var.constant for var in tensor_vars)
        if not is_const:
            # record that a variable participated in that op
            for var in tensor_vars:
                if not var.constant:
                    var._ops.append(f)
        return cls(op_out, constant=is_const, _creator=f)

    def backward(self, grad=None, terminal_node=False):
        if self.constant:
            return
        grad = np.asarray(grad) if grad is not None else np.asarray(1., dtype=float)
        if terminal_node:
            self.grad = np.asarray(grad)
        else:
            if not terminal_node and not self._ops:
                raise Exception("Invalid Backprop: part of the computational graph containing "
                                "this tensor was cleared prior to backprop")
            self.grad = np.asarray(grad if self.grad is None else self.grad + grad)

        if self.creator is not None:
            self.creator.backward(grad)

    def null_gradients(self, clear_graph=True):
        self.grad = None
        if self.creator is not None:
            self.creator.null_gradients(False)
        if clear_graph:
            self.clear_graph()

    def clear_graph(self, terminal_node=True):
        if not terminal_node:
            self._ops = []
        if self.creator is not None:
            self.creator.clear_graph(terminal_node=False)
            self.creator = None


class Multiply(SimpleOperation):
    def __call__(self, a, b):
        self.variables = (a, b)
        out = a.data * b.data
        return out

    def backward_var(self, grad, index):
        a, b = self.variables
        if index == 0:  # backprop through a
            a.backward(grad * b.data)
        elif index == 1:  # backprop through b
            b.backward(grad * a.data)


def _multiply(a, b, constant=False):
    return Node._op(Multiply, a, b, constant=constant)


class Add(SimpleOperation):
    def __call__(self, a, b):
        self.variables = (a, b)
        out = a.data + b.data
        return out

    def backward_var(self, grad, index):
        self.variables[index].backward(np.copy(grad))


def _add(a, b, constant=False):
    return Node._op(Add, a, b, constant=constant)



from ...wrappers.uber import fwdprop_test_factory as fwdprop_test_factory
from ...wrappers.uber import backprop_test_factory as backprop_test_factory
from ...custom_strategies import valid_axes
from pytest import raises

from mygrad import amax, amin, sum, mean, cumprod, cumsum, prod, var, std
import mygrad

import numpy as np

from hypothesis import settings
import hypothesis.strategies as st
import hypothesis.extra.numpy as hnp


def axis_arg(*arrs):
    """ Wrapper for passing valid-axis search strategy to test factory"""
    return valid_axes(arrs[0].ndim)


def single_axis_arg(*arrs):
    """ Wrapper for passing valid-axis (single-value only)
    search strategy to test factory"""
    return valid_axes(arrs[0].ndim, single_axis_only=True)


def keepdims_arg(*arrs):
    """ Wrapper for passing keep-dims strategy to test factory"""
    return st.booleans()


def ddof_arg(*arrs):
    """ Wrapper for passing ddof strategy to test factory
    (argument for var and std)"""
    return st.integers(0, min(arrs[0].shape) - 1)


@fwdprop_test_factory(mygrad_func=amax, true_func=np.amax, num_arrays=1,
                      kwargs=dict(axis=axis_arg, keepdims=keepdims_arg))
def test_max_fwd():
    pass


@backprop_test_factory(mygrad_func=amax, true_func=np.amax, num_arrays=1,
                       kwargs=dict(axis=axis_arg, keepdims=keepdims_arg),
                       vary_each_element=True, index_to_unique={0: True},
                       elements_strategy=st.integers)
def test_max_bkwd():
    pass


@fwdprop_test_factory(mygrad_func=amin, true_func=np.amin, num_arrays=1,
                      kwargs=dict(axis=axis_arg, keepdims=keepdims_arg))
def test_min_fwd():
    pass


@backprop_test_factory(mygrad_func=amin, true_func=np.amin, num_arrays=1,
                       kwargs=dict(axis=axis_arg, keepdims=keepdims_arg),
                       vary_each_element=True, index_to_unique={0: True},
                       elements_strategy=st.integers)
def test_min_bkwd():
    pass


def test_min_max_aliases():
    assert mygrad.max == amax
    assert mygrad.min == amin


@fwdprop_test_factory(mygrad_func=sum, true_func=np.sum, num_arrays=1,
                      kwargs=dict(axis=axis_arg, keepdims=keepdims_arg))
def test_sum_fwd():
    pass


@backprop_test_factory(mygrad_func=sum, true_func=np.sum, num_arrays=1,
                       kwargs=dict(axis=axis_arg, keepdims=keepdims_arg),
                       vary_each_element=True, atol=1e-5)
def test_sum_bkwd():
    pass


@fwdprop_test_factory(mygrad_func=mean, true_func=np.mean, num_arrays=1,
                      kwargs=dict(axis=axis_arg, keepdims=keepdims_arg))
def test_mean_fwd():
    pass


@backprop_test_factory(mygrad_func=mean, true_func=np.mean, num_arrays=1,
                       kwargs=dict(axis=axis_arg, keepdims=keepdims_arg),
                       index_to_bnds={0: (-10, 10)},
                       vary_each_element=True)
def test_mean_bkwd():
    pass


@fwdprop_test_factory(mygrad_func=var, true_func=np.var, num_arrays=1,
                      kwargs=dict(axis=axis_arg, keepdims=keepdims_arg,
                                  ddof=ddof_arg))
def test_var_fwd():
    pass


@backprop_test_factory(mygrad_func=var, true_func=np.var, num_arrays=1,
                       kwargs=dict(axis=axis_arg, keepdims=keepdims_arg,
                                   ddof=ddof_arg),
                       vary_each_element=True, index_to_bnds={0: (-10, 10)},
                       atol=1e-5, rtol=1e-5)
def test_var_bkwd():
    pass


# std composes mygrad's sqrt and var, backprop need not be tested
@fwdprop_test_factory(mygrad_func=std, true_func=np.std, num_arrays=1,
                      kwargs=dict(axis=axis_arg, keepdims=keepdims_arg,
                                  ddof=ddof_arg))
def test_std_fwd():
    pass


def _assume(*arrs, **kwargs):
    return all(i > 1 for i in arrs[0].shape)


@backprop_test_factory(mygrad_func=std, true_func=np.std, num_arrays=1,
                       kwargs=dict(axis=axis_arg, keepdims=keepdims_arg,
                                   ddof=ddof_arg),
                       vary_each_element=True, index_to_bnds={0: (-10, 10)},
                       elements_strategy=st.integers,
                       index_to_unique={0: True},
                       assumptions=_assume,
                       atol=1e-5, rtol=1e-5)
def test_std_bkwd():
    pass


@fwdprop_test_factory(mygrad_func=prod, true_func=np.prod, num_arrays=1,
                      kwargs=dict(axis=axis_arg, keepdims=keepdims_arg))
def test_prod_fwd():
    pass


@backprop_test_factory(mygrad_func=prod, true_func=np.prod, num_arrays=1,
                       kwargs=dict(axis=axis_arg, keepdims=keepdims_arg),
                       vary_each_element=True, index_to_bnds={0: (-2, 2)})
def test_prod_bkwd():
    pass


def test_int_axis_cumprod():
    """check if numpy cumprod begins to support tuples for the axis argument"""

    x = np.array([[1, 1, 0],
                  [1, 1, 0],
                  [1, 1, 0]])
    "`np.cumprod` is expected to raise a TypeError "
    "when it is provided a tuple of axes."
    with raises(TypeError):
        np.cumprod(x, axis=(0, 1))

    "`mygrad.cumprod` is expected to raise a TypeError "
    "when it is provided a tuple of axes."
    with raises(TypeError):
        cumprod(x, axis=(0, 1))


@fwdprop_test_factory(mygrad_func=cumprod, true_func=np.cumprod, num_arrays=1,
                      kwargs=dict(axis=single_axis_arg))
def test_cumprod_fwd(): pass


@settings(deadline=None)
@backprop_test_factory(mygrad_func=cumprod, true_func=np.cumprod, num_arrays=1,
                       kwargs=dict(axis=single_axis_arg),
                       vary_each_element=True, index_to_bnds={0: (-2, 2)})
def test_cumprod_bkwd():
    pass


@settings(deadline=None)
@backprop_test_factory(mygrad_func=cumprod, true_func=np.cumprod, num_arrays=1,
                       kwargs=dict(axis=single_axis_arg),
                       vary_each_element=True, index_to_bnds={0: (-.5, .5)},
                       index_to_unique={0: True},
                       index_to_arr_shapes={0: hnp.array_shapes(max_side=5, max_dims=4)})
def test_cumprod_bkwd2():
    pass


def test_int_axis_cumsum():
    """check if numpy cumsum begins to support tuples for the axis argument"""

    x = np.array([[1, 1, 0],
                  [1, 1, 0],
                  [1, 1, 0]])
    "`np.cumsum` is expected to raise a TypeError "
    "when it is provided a tuple of axes."
    with raises(TypeError):
        np.cumsum(x, axis=(0, 1))

    "`mygrad.cumsum` is expected to raise a TypeError "
    "when it is provided a tuple of axes."
    with raises(TypeError):
        cumsum(x, axis=(0, 1))


@fwdprop_test_factory(mygrad_func=cumsum, true_func=np.cumsum, num_arrays=1,
                      kwargs=dict(axis=single_axis_arg))
def test_cumsum_fwd():
    pass


@settings(deadline=None)
@backprop_test_factory(mygrad_func=cumsum, true_func=np.cumsum, num_arrays=1,
                       kwargs=dict(axis=single_axis_arg),
                       vary_each_element=True, index_to_bnds={0: (-2, 2)},
                       atol=1e-5)
def test_cumsum_bkwd():
    pass




from mygrad import log, log2, log10, log1p
from tests.wrappers.uber import backprop_test_factory, fwdprop_test_factory

import numpy as np


@fwdprop_test_factory(mygrad_func=log, true_func=np.log, index_to_bnds={0: (1e-5, 100)}, num_arrays=1)
def test_log_fwd():
    pass


@backprop_test_factory(mygrad_func=log, true_func=np.log, index_to_bnds={0: (1e-5, 100)}, num_arrays=1,
                       as_decimal=False)
def test_log_backward():
    pass


@fwdprop_test_factory(mygrad_func=log2, true_func=np.log2, index_to_bnds={0: (1e-5, 100)}, num_arrays=1)
def test_log2_fwd():
    pass


@backprop_test_factory(mygrad_func=log2, true_func=np.log2, index_to_bnds={0: (1e-5, 100)}, num_arrays=1,
                       as_decimal=False)
def test_log2_backward():
    pass


@fwdprop_test_factory(mygrad_func=log10, true_func=np.log10, index_to_bnds={0: (1e-5, 100)}, num_arrays=1)
def test_log10_fwd():
    pass


@backprop_test_factory(mygrad_func=log10, true_func=np.log10, index_to_bnds={0: (1e-5, 100)}, num_arrays=1,
                       as_decimal=False)
def test_log10_backward():
    pass


@fwdprop_test_factory(mygrad_func=log1p, true_func=np.log1p, index_to_bnds={0: (1e-5, 100)}, num_arrays=1)
def test_log1p_fwd():
    pass


@backprop_test_factory(mygrad_func=log1p, true_func=np.log1p, index_to_bnds={0: (1e-5, 100)}, num_arrays=1,
                       as_decimal=False)
def test_log1p_backward():
    pass



from tests.custom_strategies import valid_axes

from mygrad.math._special import logsumexp
from scipy import special

from hypothesis import given
import hypothesis.extra.numpy as hnp
import hypothesis.strategies as st

import numpy as np

from numpy.testing import assert_array_equal


@given(data=st.data(),
       x=hnp.arrays(shape=hnp.array_shapes(), dtype=np.float,
                    elements=st.floats(-1e6, 1e6)),
       keepdims=st.booleans())
def test_logsumexp(data: st.SearchStrategy, x: np.ndarray, keepdims: bool):
    axes = data.draw(valid_axes(ndim=x.ndim), label="axes")
    mygrad_result = logsumexp(x, axis=axes, keepdims=keepdims)
    scipy_result = special.logsumexp(x, axis=axes, keepdims=keepdims)
    assert_array_equal(mygrad_result, scipy_result, err_msg="mygrad's implementation of logsumexp does "
                                                            "not match that of scipy's")



from itertools import permutations
from mygrad.tensor_base import Tensor
from mygrad import transpose, moveaxis, swapaxes, squeeze, expand_dims, ravel, broadcast_to
from numpy.testing import assert_allclose
import numpy as np

from .custom_strategies import valid_axes, broadcastable_shape
from .utils.numerical_gradient import numerical_gradient_full
from .wrappers.uber import fwdprop_test_factory, backprop_test_factory

import hypothesis.extra.numpy as hnp
from hypothesis import given, assume, settings
import hypothesis.strategies as st

from pytest import raises
    

@settings(deadline=None)
@given(x=hnp.arrays(shape=hnp.array_shapes(max_side=4, max_dims=5),
                    dtype=float,
                    elements=st.floats(-10., 10.)),
       data=st.data())
def test_transpose(x, data):
    axes = data.draw(valid_axes(x.ndim), label="axes")
    if axes is not None:
        assume(len(axes) == x.ndim)

    x_arr = Tensor(np.copy(x))

    o = transpose(x_arr, axes, constant=False)
    grad = data.draw(hnp.arrays(shape=o.shape,
                                dtype=float,
                                elements=st.floats(1, 10),
                                unique=True),
                     label="grad")

    o.backward(grad)

    def f(x): return np.transpose(x, axes)

    assert_allclose(o.data, f(x))

    dx, = numerical_gradient_full(f, x, back_grad=grad, as_decimal=True)

    assert_allclose(x_arr.grad, dx)

    out = transpose(x, constant=True)
    assert out.constant and not x_arr.constant


def test_transpose_property():
    dat = np.arange(6).reshape(2, 3)
    x = Tensor(dat)
    f = x.T
    f.backward(dat.T)

    assert_allclose(f.data, dat.T)
    assert_allclose(x.grad, dat)


def test_transpose_method():
    dat = np.arange(24).reshape(2, 3, 4)

    for axes in permutations(range(3)):
        # passing tuple of integers
        x = Tensor(dat)
        f = x.transpose(axes)
        f.backward(dat.transpose(axes))

        assert_allclose(f.data, dat.transpose(axes))
        assert_allclose(x.grad, dat)

        # passing integers directly
        x = Tensor(dat)
        f = x.transpose(*axes)
        f.backward(dat.transpose(axes))

        assert_allclose(f.data, dat.transpose(axes), err_msg="{}".format(axes))
        assert_allclose(x.grad, dat, err_msg="{}".format(axes))

    # passing integers directly
    x = Tensor(dat)
    f = x.transpose()
    f.backward(dat.transpose())

    assert_allclose(f.data, dat.transpose())
    assert_allclose(x.grad, dat)

    # check that constant=True works
    x = Tensor(dat)
    f = x.transpose(constant=True)
    assert f.constant and not x.constant

    f = x.transpose(1, 0, 2, constant=True)
    assert f.constant and not x.constant


@given(x=hnp.arrays(shape=hnp.array_shapes(max_side=2, max_dims=3),
                    dtype=float,
                    elements=st.floats(-10., 10.)),
       data=st.data())
def test_squeeze(x, data):
    axes = data.draw(valid_axes(x.ndim), label="axes")
    x_arr = Tensor(np.copy(x))
    x_arr2 = Tensor(np.copy(x))

    def f(x): return np.squeeze(x, axes)

    try:
        numpy_out = np.squeeze(x, axes)
    except ValueError as e:
        with raises(ValueError):
            squeeze(x_arr, axes, constant=False)
        return

    o = squeeze(x_arr, axes, constant=False)
    o_method = x_arr2.squeeze(axes)
    assert_allclose(o.data, numpy_out)
    assert_allclose(o_method.data, numpy_out)

    grad = data.draw(hnp.arrays(shape=o.shape,
                                dtype=float,
                                elements=st.floats(1, 10),
                                unique=True),
                     label="grad")
    o.backward(grad)
    o_method.backward(grad)

    dx, = numerical_gradient_full(f, x, back_grad=grad, as_decimal=True)

    assert_allclose(x_arr.grad, dx)
    assert_allclose(x_arr2.grad, dx)


def _expand_dims_axis(arr):
    return st.integers(-arr.ndim - 1, arr.ndim)


def _swap_axes_axis(arr):
    return st.integers(-arr.ndim, arr.ndim - 1)


def _valid_moveaxis_args(*arrs, **kwargs):
    return len(kwargs["source"]) == len(kwargs["destination"])


@fwdprop_test_factory(mygrad_func=expand_dims, true_func=np.expand_dims,
                      num_arrays=1, kwargs=dict(axis=_expand_dims_axis))
def test_expand_dims_fwd():
    pass


@backprop_test_factory(mygrad_func=expand_dims, true_func=np.expand_dims,
                       num_arrays=1, kwargs=dict(axis=_expand_dims_axis),
                       vary_each_element=True)
def test_expand_dims_bkwd():
    pass


@fwdprop_test_factory(mygrad_func=moveaxis, true_func=np.moveaxis,
                      num_arrays=1, kwargs=dict(source=lambda x: valid_axes(x.ndim, permit_none=False),
                                                destination=lambda x: valid_axes(x.ndim, permit_none=False)),
                      assumptions=_valid_moveaxis_args,
                      index_to_arr_shapes={0: hnp.array_shapes(max_side=4, max_dims=5)})
def test_moveaxis_fwd():
    pass


@settings(deadline=None)
@backprop_test_factory(mygrad_func=moveaxis, true_func=np.moveaxis,
                       num_arrays=1, kwargs=dict(source=lambda x: valid_axes(x.ndim, permit_none=False),
                                                 destination=lambda x: valid_axes(x.ndim, permit_none=False)),
                       assumptions=_valid_moveaxis_args,
                       vary_each_element=True,
                       index_to_arr_shapes={0: hnp.array_shapes(max_side=4, max_dims=5)})
def test_moveaxis_bkwd():
    pass


@fwdprop_test_factory(mygrad_func=swapaxes, true_func=np.swapaxes,
                      num_arrays=1, kwargs=dict(axis1=_swap_axes_axis, axis2=_swap_axes_axis))
def test_swapaxes_fwd():
    pass


@backprop_test_factory(mygrad_func=swapaxes, true_func=np.swapaxes,
                       num_arrays=1, kwargs=dict(axis1=_swap_axes_axis, axis2=_swap_axes_axis),
                       vary_each_element=True)
def test_swapaxes_bkwd():
    pass


@fwdprop_test_factory(mygrad_func=Tensor.flatten, true_func=np.ndarray.flatten,
                      num_arrays=1)
def test_flatten_fwd():
    pass


@backprop_test_factory(mygrad_func=Tensor.flatten, true_func=np.ndarray.flatten,
                       num_arrays=1, vary_each_element=True)
def test_flatten_bkwd():
    pass


@fwdprop_test_factory(mygrad_func=ravel, true_func=np.ravel,
                      num_arrays=1)
def test_ravel_fwd():
    pass


@backprop_test_factory(mygrad_func=ravel, true_func=np.ravel,
                       num_arrays=1, vary_each_element=True)
def test_ravel_bkwd():
    pass


def _is_broadcastable(*arrs, **kwargs):
    x = arrs[0]
    return x.shape == kwargs["shape"][-x.ndim:]


@fwdprop_test_factory(mygrad_func=broadcast_to, true_func=np.broadcast_to,
                      num_arrays=1, kwargs=dict(shape=lambda arr: broadcastable_shape(arr.shape, min_dim=arr.ndim+1)),
                      assumptions=_is_broadcastable)
def test_broadcast_to_fwd():
    pass


@backprop_test_factory(mygrad_func=broadcast_to, true_func=np.broadcast_to,
                       num_arrays=1, vary_each_element=True,
                       kwargs=dict(shape=lambda arr: broadcastable_shape(arr.shape, min_dim=arr.ndim+1)),
                       assumptions=_is_broadcastable)
def test_broadcast_to_bkwd():
    pass



from mygrad.operation_base import Operation
import numpy as np

__all__ = ["Sinh",
           "Cosh",
           "Tanh",
           "Csch",
           "Sech",
           "Coth",
           ""
           "Arcsinh",
           "Arccosh",
           "Arctanh",
           "Arccsch",
           "Arccoth"]


class Sinh(Operation):
    """ f(a) -> sinh(a) """
    def __call__(self, a):
        self.variables = (a,)
        return np.sinh(a.data)

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        return grad * np.cosh(a.data)


class Cosh(Operation):
    """ f(a) -> cosh(a) """
    def __call__(self, a):
        self.variables = (a,)
        return np.cosh(a.data)

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        return grad * np.sinh(a.data)


class Tanh(Operation):
    """ f(a) -> tanh(a) """
    def __call__(self, a):
        self.variables = (a,)
        return np.tanh(a.data)

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        return grad * (1 - np.tanh(a.data) ** 2)


class Csch(Operation):
    """ f(a) -> csch(a) """
    def __call__(self, a):
        self.variables = (a,)
        return 1 / np.sinh(a.data)

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        return grad * -np.cosh(a.data) / np.sinh(a.data)**2


class Sech(Operation):
    """ f(a) -> sech(a) """
    def __call__(self, a):
        self.variables = (a,)
        return 1 / np.cosh(a.data)

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        return grad * -np.sinh(a.data) / np.cosh(a.data)**2


class Coth(Operation):
    """ f(a) -> coth(a) """
    def __call__(self, a):
        self.variables = (a,)
        return 1 / np.tanh(a.data)

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        return grad * -1 / np.sinh(a.data)**2


class Arcsinh(Operation):
    """ f(a) -> arcsinh(a) """
    def __call__(self, a):
        self.variables = (a,)
        return np.arcsinh(a.data)

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        return grad / np.sqrt(1 + a.data ** 2)


class Arccosh(Operation):
    """ f(a) -> arccosh(a) """
    def __call__(self, a):
        self.variables = (a,)
        return np.arccosh(a.data)

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        return grad / np.sqrt(a.data ** 2 - 1)


class Arctanh(Operation):
    """ f(a) -> arctanh(a) """
    def __call__(self, a):
        self.variables = (a,)
        return np.arctanh(a.data)

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        return grad / (1 - a.data ** 2)


class Arccsch(Operation):
    """ f(a) -> arccsch(a) """
    def __call__(self, a):
        self.variables = (a,)
        return np.arcsinh(1 / a.data)

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        return -grad / (np.abs(a.data) * np.sqrt(1 + a.data ** 2))


class Arccoth(Operation):
    """ f(a) -> arccoth(a) """
    def __call__(self, a):
        self.variables = (a,)
        return np.arctanh(1 / a.data)

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        return grad / (1 - a.data ** 2)



from ..utils.numerical_gradient import numerical_gradient_full, numerical_gradient
from ..custom_strategies import broadcastable_shape

from mygrad import Tensor

from copy import copy

from hypothesis import given, assume
import hypothesis.strategies as st
import hypothesis.extra.numpy as hnp

import numpy as np
from numpy.testing import assert_allclose, assert_array_equal
from functools import wraps
from itertools import combinations


from typing import Callable, Dict, Optional, Tuple, Sequence, Union, Any
from hypothesis.strategies import SearchStrategy


class fwdprop_test_factory():
    """ Decorator

        Randomly draw N arrays x, (...) to verify that a mygrad function,
        `f(x, ..., **kwargs)` returns a correct result through forward-propagation.

        By default, all arrays will have shapes that are broadcast-compatible with x.

        This decorator is extremely uber: the decorated function's body is
        never executed; this decorator builds the entire unit tests The
        function definition is effectively there to name the test being
        constructed.

        The user specifies the number of arrays to be generated (at least one must be
        generated), along with the bounds on their elements, their shapes, as well
        as the keyword-args to be passed to the function.

        Examples
        --------
        Writing a test that compares mygrad's add to numpy's add

        >>> from mygrad import add
        >>> import numpy as np
        >>> @fwdprop_test_factory(mygrad_func=add, true_func=np.add, num_array_args=2)
        ... def test_add():
        ...     pass
        """
    def __init__(self, *,
                 mygrad_func: Callable[[Tensor], Tensor],
                 true_func: Callable[[np.ndarray], np.ndarray],
                 num_arrays: int,
                 index_to_bnds: Optional[Dict[int, Tuple[int, int]]] = {},
                 index_to_no_go: Optional[Dict[int, Sequence[int]]] = {},
                 kwargs: Optional[Dict[str, Union[Any, Callable[[Any], SearchStrategy]]]] = {},
                 index_to_arr_shapes: Optional[Dict[int, Union[Sequence[int], SearchStrategy]]] = {},
                 assumptions: Optional[Callable[..., bool]] = None):
        """
        Parameters
        ----------
        mygrad_func : Callable[[numpy.ndarray, ...], mygrad.Tensor]
            The mygrad function whose forward pass validity is being checked.

        true_func : Callable[[numpy.ndarray, ...], numpy.ndarray]
            A known correct version of the function

        index_to_bnds : Dict[int, Tuple[int, int]]
            Indicate the lower and upper bounds from which the elements
            for array-i is drawn. By default, [-10, 10].

        index_to_no_go : Dict[int, Sequence[int]]
            Values that array-i cannot possess. By default, no values are
            excluded.

        index_to_arr_shapes : Dict[int, Union[Sequence[int], hypothesis.searchstrategy.SearchStrategy]]
            The shape for array-i. This can be an exact shape or a hypothesis search
            strategy that draws shapes.
                Default for array-0: `hnp.array_shapes(max_side=3, max_dims=3)`
                Default for array-i: `broadcastable_shape(arr-0.shape)`

        kwargs : Dict[str, Union[Any, Callable[[Any], hypothesis.searchstrategy.SearchStrategy]]]
            Keyword arguments and their values to be passed to the functions.
            The values can be hypothesis search-strategies, in which case
            a value when be drawn at test time for that argument using the provided
            strategy.

            Note that any search strategy must be "wrapped" in a function, which
            will be called, passing it the list of arrays as an input argument, such
            that the strategy can draw based on those particular arrays.

        assumptions : Optional[Callable[[arrs, **kwargs], bool]]
            A callable that is fed the generated arrays and keyword arguments that will
            be fed to ``mygrad_func``. If ``assumptions`` returns ``False``, that test
            case will be marked as skipped by hypothesis.
        """
        assert num_arrays > 0
        self.op = mygrad_func
        self.true_func = true_func

        self.index_to_bnds = index_to_bnds
        self.index_to_no_go = index_to_no_go
        self.index_to_arr_shapes = index_to_arr_shapes
        self.kwargs = kwargs
        self.num_arrays = num_arrays
        self.assumptions = assumptions

    def gen_first_array(self) -> st.SearchStrategy:
        """
        Hypothesis search strategy for drawing the array x to be passed to f(x, ...)

        Returns
        -------
        hypothesis.searchstrategy.SearchStrategy"""
        return hnp.arrays(shape=self.index_to_arr_shapes.get(0, hnp.array_shapes(max_side=3, max_dims=3)),
                          dtype=float,
                          elements=st.floats(*self.index_to_bnds.get(0, (-10., 10.))))

    def gen_other_array(self, x: np.ndarray, i: int) -> st.SearchStrategy:
        """
        Hypothesis search strategy for drawing an array y to be passed to f(x, ..., y_i,...).
        By default, y is drawn to have a shape that is broadcast-compatible with x.

        Parameters
        ----------
        x : numpy.ndarray
        i : int
            The argument index-location of y in the signature of f.

        Returns
        -------
        hypothesis.searchstrategy.SearchStrategy"""
        return hnp.arrays(shape=self.index_to_arr_shapes.get(i, broadcastable_shape(x.shape)),
                          dtype=float,
                          elements=st.floats(*self.index_to_bnds.get(i, (-10., 10.))))

    def __call__(self, f):
        @given(x=self.gen_first_array(),
               constant=st.booleans(),
               data=st.data())
        @wraps(f)
        def wrapper(x, constant, data):
            arrs = [x]  # list of drawn arrays to feed to functions

            for i in range(1, self.num_arrays):  # draw additional arrays according to `num_arrays`
                y = data.draw(self.gen_other_array(x, i), label="array-{}".format(i))
                arrs.append(y)

            arr_copies = [copy(arr) for arr in arrs]  # list of array-copies to check for mutation

            # set or draw keyword args to be passed to functions
            kwargs = {k: (data.draw(v(*arrs), label="kwarg: {}".format(k)) if callable(v) else v)
                      for k, v in self.kwargs.items()}

            if self.assumptions is not None:
                assume(self.assumptions(*arrs, **kwargs))

            for i, arr in enumerate(arrs):  # assure arrays don't contain forbidden values
                for value in self.index_to_no_go.get(i, ()):
                    assume(np.all(arr != value))

            # execute mygrad and "true" functions. Compare outputs and check mygrad behavior
            o = self.op(*(Tensor(i) for i in arrs), **kwargs, constant=constant)
            tensor_out = o.data
            true_out = self.true_func(*arrs, **kwargs)

            assert isinstance(o, Tensor), \
                "`mygrad_func` returned type {}, should return `mygrad.Tensor`".format(type(o))
            assert o.constant is constant, \
                "`mygrad_func` returned tensor.constant={}, should be constant={}".format(o.constant, constant)

            assert_allclose(actual=tensor_out, desired=true_out,
                            err_msg="`mygrad_func(x)` and `true_func(x)` produce different results")

            for n, (arr, arr_copy) in enumerate(zip(arrs, arr_copies)):
                assert_array_equal(arr, arr_copy,
                                   err_msg="arr-{} was mutated during forward prop".format(n))
        return wrapper


class backprop_test_factory():
    """ Decorator

        Randomly draw arrays x, ... to verify that a binary mygrad function,
        `f(x, ..., **kwargs)` performs back-propagation appropriately.

        x.grad, ... are compared against numerical derivatives of f.

        This decorator is extremely uber: the decorated function's body is
        never executed. The function definition is effectively there to name
        the test being constructed. This constructs the entire test

        Examples
        --------
        >>> from mygrad import add
        >>> import numpy as np
        >>> @backprop_test_factory(mygrad_func=add, true_func=np.add)
        ... def test_add():
        ...     pass"""

    def __init__(self, *,
                 mygrad_func: Callable[[Tensor], Tensor],
                 true_func:  Callable[[np.ndarray], np.ndarray],
                 num_arrays: int,
                 index_to_bnds: Optional[Dict[int, Tuple[int, int]]]=None,
                 index_to_no_go: Optional[Dict[int, Sequence[int]]]=None,
                 index_to_arr_shapes: Optional[Dict[int, Union[Sequence[int], SearchStrategy]]]=None,
                 index_to_unique: Optional[Union[Dict[int, bool], bool]]=None,
                 elements_strategy: Optional[SearchStrategy]=None,
                 kwargs: Optional[Dict[str, Union[Any, Callable[[Any], SearchStrategy]]]]=None,
                 h: float=1e-8,
                 rtol: float=1e-05,
                 atol: float=1e-05,
                 vary_each_element: bool=False,
                 as_decimal: bool=True,
                 assumptions: Optional[Callable[..., bool]] = None):
        """
        Parameters
        ----------
        mygrad_func : Callable[[numpy.ndarray, ...], mygrad.Tensor]
            The mygrad function whose backward pass validity is being checked.

        true_func : Callable[[numpy.ndarray, ...], numpy.ndarray]
            A known correct version of the function, which is used to compute
            numerical derivatives.

        num_arrays : int
            The number of arrays that must be passed to ``mygrad_func``

        index_to_bnds : Optional[Dict[int, Tuple[int, int]]]
            Indicate the lower and upper bounds from which the elements
            for array-i is drawn. By default, [-100, 100].

        index_to_no_go : Optional[Dict[int, Sequence[int]]]
            Values that array-i cannot possess. By default, no values are
            excluded.

        index_to_arr_shapes : Optional[Dict[int, Union[Sequence[int], SearchStrategy]]]
            The shape for array-i. This can be an exact shape or a hypothesis search
            strategy that draws shapes.
                Default for array-0: `hnp.array_shapes(max_side=3, max_dims=3)`
                Default for array-i: `broadcastable_shape(arr-0.shape)`

        index_to_unique : Optional[Union[Dict[int, bool], bool]]
            Determines whether the elements drawn for each of the input-arrays are
            required to be unique or not. By default this is `False` for each array.
            If a single boolean value is supplied, this is applied for every array.

        elements_strategy : Optional[Union[SearchStrategy]
            The hypothesis-type-strategy used to draw the array elements.
            The default value is ``hypothesis.strategies.floats``.

        kwargs : Optional[Dict[str, Union[Any, Callable[[Any], SearchStrategy]]]]
            Keyword arguments and their values to be passed to the functions.
            The values can be hypothesis search strategies, in which case
            a value when be drawn at test time for that argument.

            Note that any search strategy must be "wrapped" in a function, which
            will be called, passing it the list of arrays as an input argument, such
            that the strategy can draw based on those particular arrays.

        vary_each_element : bool, optional (default=False)
            If False, then use a faster numerical derivative that varies entire
            arrays at once: arr -> arr + h; valid only for functions that map over
            entries, like 'add' and 'sum'. Otherwise, the gradient is constructed
            by varying each element of each array independently.

        as_decimal : bool, optional (default=True)
            If True, x is passed to f as a Decimal-type array. This
            improves numerical precision, but is not permitted by some functions.

        assumptions : Optional[Callable[[arrs, **kwargs], bool]]
            A callable that is fed the generated arrays and keyword arguments that will
            be fed to ``mygrad_func``. If ``assumptions`` returns ``False``, that test
            case will be marked as skipped by hypothesis.
        """

        index_to_bnds = index_to_bnds if index_to_bnds is not None else {}
        index_to_no_go = index_to_no_go if index_to_no_go is not None else {}
        index_to_arr_shapes = index_to_arr_shapes if index_to_arr_shapes is not None else {}
        index_to_unique = index_to_unique if index_to_unique is not None else {}
        self.elements_strategy = elements_strategy if elements_strategy is not None else st.floats
        kwargs = kwargs if kwargs is not None else {}

        assert num_arrays > 0
        self.op = mygrad_func
        self.true_func = true_func

        if isinstance(index_to_bnds, (tuple, list, np.ndarray)):
            index_to_bnds = {k: index_to_bnds for k in range(num_arrays)}
        self.index_to_bnds = index_to_bnds

        if isinstance(index_to_no_go, (tuple, list, np.ndarray)):
            index_to_no_go = {k: index_to_no_go for k in range(num_arrays)}
        self.index_to_no_go = index_to_no_go

        if isinstance(index_to_arr_shapes, (tuple, list, np.ndarray, st.SearchStrategy)):
            index_to_arr_shapes = {k: index_to_arr_shapes for k in range(num_arrays)}
            self.index_to_arr_shapes = index_to_arr_shapes
        self.index_to_arr_shapes = index_to_arr_shapes

        if isinstance(index_to_unique, bool):
            index_to_unique = {k: index_to_unique for k in range(num_arrays)}
        self.index_to_unique = index_to_unique
        self.kwargs = kwargs
        self.num_arrays = num_arrays

        self.h = h
        self.tolerances = dict(rtol=rtol, atol=atol)
        self.vary_each_element = vary_each_element
        self.as_decimal = as_decimal
        self.assumptions = assumptions

    def gen_first_array(self) -> st.SearchStrategy:
        """
        Hypothesis search strategy for drawing the array x to be passed to f(x, ...)

        Returns
        -------
        hypothesis.searchstrategy.SearchStrategy"""
        return hnp.arrays(shape=self.index_to_arr_shapes.get(0, hnp.array_shapes(max_side=3, max_dims=3)),
                          dtype=float,
                          elements=self.elements_strategy(*self.index_to_bnds.get(0, (-100, 100))),
                          unique=self.index_to_unique.get(0, False))

    def gen_other_array(self, x: np.ndarray, i: int) -> st.SearchStrategy:
        """
        Hypothesis search strategy for drawing an array y to be passed to f(x, ..., y_i,...).
        By default, y is drawn to have a shape that is broadcast-compatible with x.

        Parameters
        ----------
        x : numpy.ndarray
        i : int
            The argument index-location of y in the signature of f.

        Returns
        -------
        hypothesis.searchstrategy.SearchStrategy"""
        return hnp.arrays(shape=self.index_to_arr_shapes.get(i, broadcastable_shape(x.shape)),
                          dtype=float,
                          elements=self.elements_strategy(*self.index_to_bnds.get(i, (-100, 100))),
                          unique=self.index_to_unique.get(i, False))

    def __call__(self, f):
        @given(x=self.gen_first_array(),
               data=st.data())
        @wraps(f)
        def wrapper(x, data):
            arrs = [x]  # list of drawn arrays to feed to functions
            # draw additional arrays according to `num_arrays`
            for i in range(1, self.num_arrays):
                arrs.append(data.draw(self.gen_other_array(x, i), label="array-{}".format(i)))

            arrs = tuple(Tensor(arr) for arr in arrs)
            arr_copies = tuple(copy(arr) for arr in arrs)

            # The keyword args to be passed to `self.op`. If any provided argument is callable
            # it is assumed to by a hypothesis search strategy, and all of the drawn arrays will
            # be passed to the strategy, in order to draw a value for that keyword argument.
            # Otherwise the provided value is used as-is.
            kwargs = {k: (data.draw(v(*arrs), label="kwarg: {}".format(k)) if callable(v) else v)
                      for k, v in self.kwargs.items()}

            if self.assumptions is not None:
                assume(self.assumptions(*arrs, **kwargs))

            for i, arr in enumerate(arrs):  # assure arrays don't contain forbidden values
                for value in self.index_to_no_go.get(i, ()):
                    assume(np.all(arr != value))

            # forward pass of the function
            out = self.op(*arrs, **kwargs)

            # gradient to be backpropped through this operation
            grad = data.draw(hnp.arrays(shape=out.shape,
                                        dtype=float,
                                        elements=st.floats(-10, 10),
                                        unique=True),
                             label="grad")
            grad_copy = copy(grad)  # keep a copy to check for later mutations

            # compute analytic derivatives via mygrad-backprop
            if any(out.shape != i.shape for i in arrs):
                # Broadcasting occurred
                # Must reduce `out` to scalar
                # first multiply by `grad` to simulate non-trivial back-prop
                (grad * out).sum().backward()
            else:
                out.backward(grad)

            # compute derivatives via numerical approximation of derivative
            numerical_grad = numerical_gradient_full if self.vary_each_element else numerical_gradient
            grads_numerical = numerical_grad(self.true_func, *(i.data for i in arrs),
                                             back_grad=grad, kwargs=kwargs,
                                             as_decimal=self.as_decimal)

            # check that the analytic and numeric derivatives match
            for n, (arr, d_num) in enumerate(zip(arrs, grads_numerical)):
                assert_allclose(arr.grad, d_num, **self.tolerances,
                                err_msg="arr-{}: mygrad derivative and numerical derivative do not match".format(n))

                # check that none of the set derivatives is a view of `grad`
                assert not np.shares_memory(arr.grad, grad), "arr-{}.grad stores a view of grad".format(n)

            # check that none of the set derivatives are views of one another
            for arr_i, arr_j in combinations(arrs, 2):
                assert not np.shares_memory(arr_i.grad, arr_j.grad), \
                    "two input arrays were propagated views of the same gradient"

            # verify that null_gradients works
            out.null_gradients()
            assert all(i.grad is None for i in arrs), "null_gradients failed"

            # check if any of the input-arrays were mutated
            for n, (arr, arr_copy) in enumerate(zip(arrs, arr_copies)):
                assert_array_equal(arr.data, arr_copy.data,
                                   err_msg="arr-{} was mutated during backward prop".format(n))

            # check if `grad` was mutated
            assert_array_equal(grad, grad_copy,
                               err_msg="`grad` was mutated during backward prop")
        return wrapper



from .ops import Transpose, MoveAxis, SwapAxes
from mygrad.tensor_base import Tensor

__all__ = ["transpose",
           "moveaxis",
           "swapaxes"]


def transpose(a, *axes, constant=False):
    """ Permute the dimensions of a tensor.

        Parameters
        ----------
        a : array_like
            The tensor to be transposed

        axes : Optional[Tuple[int]]
            By default, reverse the dimensions, otherwise permute the axes
            according to the values given.

        Returns
        -------
        mygrad.Tensor
            `a` with its axes permuted.  A new tensor is returned. 

        Examples
        --------
        >>> import mygrad as mg
        >>> a = mg.Tensor([[1, 2], [3, 4]])
        >>> a
        Tensor([[1, 2],
                [3, 4]])
        >>> a.transpose()
        Tensor([[1, 3],
                [2, 4]])
        >>> a.transpose((1, 0))
        Tensor([[1, 3],
                [2, 4]])
        >>> a.transpose(1, 0)
        Tensor([[1, 3],
                [2, 4]]) """
    if not axes:
        axes = None
    elif hasattr(axes[0], "__iter__") or axes[0] is None:
        if len(axes) > 1:
            raise TypeError("'{}' object cannot be interpreted as an integer".format(type(axes[0])))
        axes = axes[0]
    return Tensor._op(Transpose, a, op_args=(axes,), constant=constant)


def moveaxis(a, source, destination, constant=False):
    """ Move axes of a tensor to new positions. Other axes remain in their
        original order.


        Parameters
        ----------
        a : array_like
            The array whose axes should be reordered.

        source : Union[int, Sequence[int]]
            Original positions of the axes to move. These must be unique.

        destination : Union[int, Sequence[int]]
            Destination positions for each of the original axes. These must also be
            unique.

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        result : mygrad.Tensor
            Array with moved axes. This array is a view of the input array..

        Examples
        --------
        >>> from mygrad import Tensor, moveaxis
        >>> x = Tensor(np.zeros((3, 4, 5)))
        >>> moveaxis(x, 0, -1).shape
        (4, 5, 3)
        >>> moveaxis(x, -1, 0).shape
        (5, 3, 4)
        >>> moveaxis(x, [0, 1], [-1, -2]).shape
        (5, 4, 3) """
    return Tensor._op(MoveAxis, a, op_args=(source, destination), constant=constant)


def swapaxes(a, axis1, axis2, constant=False):
    """ Interchange two axes of a tensor.

        Parameters
        ----------
        a : array_like
            Input array.

        axis1 : int
            First axis.

        axis2 : int
            Second axis.

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        mygrad.Tensor

        Examples
        --------
        >>> from mygrad import Tensor, swapaxes
        >>> x = Tensor([[1, 2, 3]])
        >>> swapaxes(x, 0, 1)
        Tensor([[1],
               [2],
               [3]])
        >>> x = Tensor([[[0, 1], [2, 3]], [[4, 5], [6, 7]]])
        >>> x
        Tensor([[[0, 1],
                [2, 3]],
               [[4, 5],
                [6, 7]]])
        >>> swapaxes(x, 0, 2)
        Tensor([[[0, 4],
                [2, 6]],
               [[1, 5],
                [3, 7]]])
    """
    return Tensor._op(SwapAxes, a, op_args=(axis1, axis2), constant=constant)



from mygrad.operation_base import Operation
from mygrad.tensor_base import Tensor
from numbers import Integral
import numpy as np

try:
    from numba import njit, vectorize
except ImportError:
    raise ImportError("The package `numba` must be installed in order to access the gru.")


@vectorize(['int32(int32)',
            'int64(int64)',
            'float32(float32)',
            'float64(float64)'], nopython=True)
def sig(f):
    """
    Calculates a sigmoid function
    """
    return 1 / (1 + np.exp(-f))


@vectorize(['int32(int32)',
            'int64(int64)',
            'float32(float32)',
            'float64(float64)'], nopython=True)
def d_sig(f):
    """
    Calculates the derivative of a sigmoid function
    """
    return f * (1 - f)


@vectorize(['int32(int32)',
            'int64(int64)',
            'float32(float32)',
            'float64(float64)'], nopython=True)
def d_tanh(f):
    """
    Calculates the derivative of a tanh function
    """
    return 1 - f ** 2


@njit
def dot(a, b):
    """
    Calculates the dot product between 2 arrays
    of shapes (W,X,Y) and (Y,Z), respectively
    """
    out = np.zeros((a.shape[0], a.shape[1], b.shape[-1]))
    for i in range(len(a)):
        out[i] = np.dot(a[i], b)
    return out


@njit
def _gru_layer(s, z, r, h, Wz, Wr, Wh):
    """ Given:
            S(t=0)
            z = X(t) Uz + bz
            r = X(t) Ur + br
            h = X(t) Uh + bh

        Compute Z(t), R(t), H(t), S(t) for all 1 <= t <= T

        Parameters
        ----------
        s : numpy.ndarray, shape=(T+1, N, D)
            Modified in-place
        z : numpy.ndarray, shape=(T, N, D)
            Modified in-place
        r : numpy.ndarray, shape=(T, N, D)
            Modified in-place
        h : numpy.ndarray, shape=(T, N, D)
            Modified in-place
        Wz : numpy.ndarray, shape=(D, D)
        Wr : numpy.ndarray, shape=(D, D)
        Wh : numpy.ndarray, shape=(D, D) """
    for n in range(len(s) - 1):
        z[n] += np.dot(s[n], Wz)
        z[n] = sig(z[n])

        r[n] += np.dot(s[n], Wr)
        r[n] = sig(r[n])

        h[n] += np.dot(r[n] * s[n], Wh)
        h[n] = np.tanh(h[n])

        s[n + 1] = (1 - z[n]) * h[n] + z[n] * s[n]


@njit
def _gru_dLds(s, z, r, dLds, Wz, Wh, Wr, dz, dh, dr, s_h, one_z):
    """
                        Z_{t} = sigmoid(Uz X_{t} + Wz S_{t-1} + bz)
                        R_{t} = sigmoid(Ur X_{t} + Wr S_{t-1} + br)
                        H_{t} = tanh(Uh X_{t} + Wh (R{t} * S_{t-1}) + bh)
                        S_{t} = (1 - Z{t}) * H{t} + Z{t} * S_{t-1}

    Returns
    --------

        dL / ds(t) =   partial dL / ds(t+1) * ds(t+1) / ds(t)
                     + partial dL / ds(t+1) * ds(t+1) / dz(t) * dz(t) / ds(t)
                     + partial dL / ds(t+1) * ds(t+1) / dh(t) * dh(t) / ds(t)
                     + partial dL / ds(t+1) * ds(t+1) / dh(t) * dh(t) / dr(t) * dr(t) / ds(t)
    """
    dLdh = dot(dLds * one_z * dh, Wh)

    out = z * dLds
    out += dot(dLds * s_h * dz, Wz)
    out += dLdh * r
    out += dot(dLdh * s * dr, Wr)

    return out


@njit
def _gru_bptt(X, dLds, s, z, r, Wz, Wh, Wr, dz, dh, dr, s_h, one_z, bp_lim, old_dLds=None):
    Wz, Wh, Wr = Wz.T, Wh.T, Wr.T
    bptt = bp_lim < len(X) - 1
    if bptt:
        old_dLds = np.zeros_like(dLds)

    for i in range(bp_lim):
        #  dL(t) / ds(t) + dL(t+1) / ds(t)
        if bptt:
            source_index = slice(1, len(dLds) - i)
            target_index = slice(None, len(dLds) - (i + 1))
            dt = dLds[source_index] - old_dLds[source_index]
            old_dLds = np.copy(dLds)
        else:  # no backprop truncation
            source_index = slice(len(dLds) - (i + 1), len(dLds) - i)
            target_index = slice(len(dLds) - (i + 2), len(dLds) - (i + 1))
            dt = dLds[source_index]

        dLds[target_index] += _gru_dLds(s[source_index],
                                        z[source_index],
                                        r[source_index],
                                        dt,
                                        Wz,
                                        Wh,
                                        Wr,
                                        dz[source_index],
                                        dh[source_index],
                                        dr[source_index],
                                        s_h[source_index],
                                        one_z[source_index])


def _backprop(var, grad):
    if not var.constant:
        if var.grad is None:
            var.grad = np.asarray(grad)
        else:
            var.grad += grad


class GRUnit(Operation):
    scalar_only = True

    def __call__(self, X, Uz, Wz, bz, Ur, Wr, br, Uh, Wh, bh, s0=None, bp_lim=None, dropout=0.):
        if bp_lim is not None:
            assert isinstance(bp_lim, Integral) and 0 <= bp_lim < len(X)
        assert 0. <= dropout < 1.
        self._dropout = dropout
        self.bp_lim = bp_lim if bp_lim is not None else len(X) - 1

        self.X = X    # type: Tensor  # shape=(T, N, C)

        self.Uz = Uz  # type: Tensor  # shape=(C, D)
        self.Wz = Wz  # type: Tensor  # shape=(D, D)
        self.bz = bz  # type: Tensor  # shape=(D,)

        self.Ur = Ur  # type: Tensor  # shape=(C, D)
        self.Wr = Wr  # type: Tensor  # shape=(D, D)
        self.br = br  # type: Tensor  # shape=(D,)

        self.Uh = Uh  # type: Tensor  # shape=(C, D)
        self.Wh = Wh  # type: Tensor  # shape=(D, D)
        self.bh = bh  # type: Tensor  # shape=(D,)

        self.variables = (self.X,
                          self.Uz, self.Wz, self.bz,
                          self.Ur, self.Wr, self.br,
                          self.Uh, self.Wh, self.bh)
        T, N, C = X.shape
        D, = bz.shape

        seq = self.X.data

        # t starts at 0 for S; all other sequences begin at t = 1
        out = np.zeros((T + 1, N, D))

        if s0 is not None:
            out[0] = s0.data if isinstance(s0, Tensor) else s0

        # compute all contributions to Z, R, H from the input sequence
        # shape: T, N, D
        z = np.tensordot(seq, self.Uz.data, [[-1], [0]])
        r = np.tensordot(seq, self.Ur.data, [[-1], [0]])
        h = np.tensordot(seq, self.Uh.data, [[-1], [0]])

        if dropout:
            p = 1 - dropout
            # For Uz/Ur/Uh: a dropout mask is generated for each datum and is applied uniformly across T
            self._dropUz, self._dropUr, self._dropUh = np.random.binomial(1, p, size=(3, 1, N, D)) / p
            self._dropWz, self._dropWr, self._dropWh = np.random.binomial(1, p, size=(3, D, D)) / p

            z *= self._dropUz
            r *= self._dropUr
            h *= self._dropUh

            Wz = self._dropWz * self.Wz.data
            Wr = self._dropWr * self.Wr.data
            Wh = self._dropWh * self.Wh.data

        else:
            self._dropUz, self._dropUr, self._dropUh = None, None, None
            self._dropWz, self._dropWr, self._dropWh = None, None, None
            Wz = self.Wz.data
            Wr = self.Wr.data
            Wh = self.Wh.data

        z += bz.data  # X Uz + bz
        r += br.data  # X Ur + br
        h += bh.data  # X Uh + bh

        _gru_layer(out, z, r, h, Wz, Wr, Wh)

        self._hidden_seq = Tensor(out, _creator=self)
        self._z = Tensor(z, _creator=self)
        self._r = Tensor(r, _creator=self)
        self._h = Tensor(h, _creator=self)

        return self._hidden_seq

    def backward(self, grad, *, graph, **kwargs):
        if all(i.constant for i in [self.X,
                                    self.Uz, self.Wz, self.bz,
                                    self.Ur, self.Wr, self.br,
                                    self.Uh, self.Wh, self.bh]):
            return None

        s = self._hidden_seq.data[:-1]
        z = self._z.data
        r = self._r.data
        h = self._h.data

        dLds = grad[1:]

        const = {"1 - h**2": d_tanh(h),
                 "z*(1 - z)": d_sig(z),
                 "r*(1 - r)": d_sig(r)}

        if self._dropout:
            Wz = self._dropWz * self.Wz.data
            Wr = self._dropWr * self.Wr.data
            Wh = self._dropWh * self.Wh.data
        else:
            Wz = self.Wz.data
            Wr = self.Wr.data
            Wh = self.Wh.data

        const["s - h"] = s - h
        const["1 - z"] = 1 - z

        _gru_bptt(self.X.data, dLds, s, z, r,
                  Wz,
                  Wh,
                  Wr,
                  const["z*(1 - z)"],
                  const["1 - h**2"],
                  const["r*(1 - r)"],
                  const["s - h"],
                  const["1 - z"],
                  self.bp_lim)

        zgrad = dLds * const["s - h"]   # dL / dz
        hgrad = dLds * const["1 - z"]   # dL / dh
        rgrad = dot(const["1 - h**2"] * hgrad, Wh.T) * s  # dL / dr

        self._hidden_seq.grad = dLds

        if any(not const for const in (self.Uz.constant, self.Wz.constant, self.bz.constant)):
            dz = zgrad * const["z*(1 - z)"]
        # backprop through Wz
        if not self.Wz.constant:
            dWz = np.tensordot(s, dz, ([0, 1], [0, 1]))
            if self._dropout:
                dWz *= self._dropWz
            _backprop(self.Wz, dWz)  # self.Wz.backward(dWz, **kwargs)
        # backprop through bz
        if not self.bz.constant:
            _backprop(self.bz, dz.sum(axis=(0, 1)))
        # backprop through bz
        if not self.Uz.constant:
            if self._dropout:
                dz *= self._dropUz  # IMPORTANT augmented update: this must come after Wz and bz backprop
            _backprop(self.Uz, np.tensordot(self.X.data, dz, ([0, 1], [0, 1])))

        if any(not const for const in (self.Ur.constant, self.Wr.constant, self.br.constant)):
            dr = rgrad * const["r*(1 - r)"]
        # backprop through Wr
        if not self.Wr.constant:
            dWr = np.tensordot(s, dr, ([0, 1], [0, 1]))
            if self._dropout:
                dWr *= self._dropWr
            _backprop(self.Wr, dWr)
        # backprop through br
        if not self.br.constant:
            _backprop(self.br, dr.sum(axis=(0, 1)))  # self.br.backward(dr.sum(axis=(0, 1)), **kwargs)
        # backprop through Ur
        if not self.Ur.constant:
            if self._dropout:
                dr *= self._dropUr  # IMPORTANT augmented update: this must come after Wr and br backprop
            _backprop(self.Ur, np.tensordot(self.X.data, dr, ([0, 1], [0, 1])))

        if any(not const for const in (self.Uh.constant, self.Wh.constant, self.bh.constant)):
            dh = hgrad * const["1 - h**2"]
        # backprop through Wh
        if not self.Wh.constant:
            dWh = np.tensordot((s * r), dh, ([0, 1], [0, 1]))
            if self._dropout:
                dWh *= self._dropWh
            _backprop(self.Wh, dWh) # self.Wh.backward(dWh, **kwargs)
        # backprop through bh
        if not self.bh.constant:
            _backprop(self.bh, dh.sum(axis=(0, 1))) # self.bh.backward(dh.sum(axis=(0, 1)), **kwargs)
        # backprop through Uh
        if not self.Uh.constant:
            if self._dropout:
                dh *= self._dropUh  # IMPORTANT augmented update: this must come after Wh and bh backprop
            _backprop(self.Uh, np.tensordot(self.X.data, dh, ([0, 1], [0, 1])))

        # backprop through X
        if not self.X.constant:
            tmp = dLds * const["1 - z"] * const["1 - h**2"]
            if not self._dropout:
                dLdX = dot((dLds * const["s - h"]) * const["z*(1 - z)"], self.Uz.data.T)
                dLdX += dot(tmp, self.Uh.data.T)
                dLdX += dot(dot(tmp, Wh.T) * s * const["r*(1 - r)"], self.Ur.data.T)
            else:
                dLdX = dot((self._dropUz * (dLds * const["s - h"]) * const["z*(1 - z)"]), self.Uz.data.T)
                dLdX += dot(self._dropUh * tmp, self.Uh.data.T)
                dLdX += dot(self._dropUr * (dot(tmp, Wh.T) * s * const["r*(1 - r)"]), self.Ur.data.T)
            _backprop(self.X, dLdX)  # self.X.backward(dLdX, **kwargs)

        del self._z
        del self._r
        del self._h

        for x in (self.X,
                  self.Uz, self.Wz, self.bz,
                  self.Ur, self.Wr, self.br,
                  self.Uh, self.Wh, self.bh):
            if not x.constant:
                x._accum_ops.add(self)
                x._backward(graph=graph)


def gru(X, Uz, Wz, bz, Ur, Wr, br, Uh, Wh, bh, s0=None, bp_lim=None, dropout=0., constant=False):
    r""" Performs a forward pass of sequential data through a Gated Recurrent Unit layer, returning
        the 'hidden-descriptors' arrived at by utilizing the trainable parameters as follows::

                    Z_{t} = sigmoid(X_{t} Uz + S_{t-1} Wz + bz)
                    R_{t} = sigmoid(X_{t} Ur + S_{t-1} Wr + br)
                    H_{t} =    tanh(X_{t} Uh + (R{t} * S_{t-1}) Wh + bh)
                    S_{t} = (1 - Z{t}) * H{t} + Z{t} * S_{t-1}

        Parameters
        ----------
        X : array_like, shape=(T, N, C)
           The sequential data to be passed forward.

        Uz : array_like, shape=(C, D)
           The weights used to map sequential data to its hidden-descriptor representation

        Wz : array_like, shape=(D, D)
            The weights used to map a hidden-descriptor to a hidden-descriptor.

        bz : array_like, shape=(D,)
           The biases used to scale a hidden-descriptor.

        Ur : array_like, shape=(C, D)
           The weights used to map sequential data to its hidden-descriptor representation

        Wr : array_like, shape=(D, D)
            The weights used to map a hidden-descriptor to a hidden-descriptor.

        br : array_like, shape=(D,)
           The biases used to scale a hidden-descriptor.

        Uh : array_like, shape=(C, D)
           The weights used to map sequential data to its hidden-descriptor representation

        Wh : array_like, shape=(D, D)
            The weights used to map a hidden-descriptor to a hidden-descriptor.

        bh : array_like, shape=(D,)
           The biases used to scale a hidden-descriptor.

        s0 : Optional[array_like], shape=(N, D)
            The 'seed' hidden descriptors to feed into the RNN. If None, a Tensor
            of zeros of shape (N, D) is created.

        bp_lim : Optional[int]
            The (non-zero) limit of the depth of back propagation through time to be
            performed. If `None` back propagation is passed back through the entire sequence.

            E.g. `bp_lim=3` will propagate gradients only up to 3 steps backward through the
            recursive sequence.

        dropout : float (default=0.), 0 <= dropout < 1
            If non-zero, the dropout scheme described in [1]_ is applied. See Notes
            for more details.

        constant : bool, optional (default=False)
            If True, the resulting Tensor is a constant.

        Returns
        -------
        mygrad.Tensor, shape=(T+1, N, D)
            The sequence of 'hidden-descriptors' produced by the forward pass of the RNN.

        Notes
        -----
        - :math:`T` : Sequence length
        - :math:`N` : Batch size
        - :math:`C` : Length of single datum
        - :math:`D` : Length of 'hidden' descriptor

        The GRU system of equations is given by:

        .. math::

                    Z_{t} = \sigma (X_{t} U_z + S_{t-1} Wz + bz)

                    R_{t} = \sigma (X_{t} U_r + S_{t-1} Wr + br)

                    H_{t} =  tanh(X_{t} U_h + (R_{t} * S_{t-1}) W_h + b_h)

                    S_{t} = (1 - Z_{t}) * H_{t} + Z_{t} * S_{t-1}

        Following the dropout scheme specified in [1]_, the hidden-hidden weights (Wz/Wr/Wh)
        randomly have their weights dropped prior to forward/back-prop. The input connections
        (via Uz/Ur/Uh) have variational dropout ([2]_) applied to them with a common dropout
        mask across all t. That is three static dropout masks, each with shape-(N,D), are
        applied to

        .. math::
                                              X_{t} U_z

                                              X_{t} U_r

                                              X_{t} U_h
        respectively, for all :math:`t`.

        References
        ----------
        .. [1] S. Merity, et. al. "Regularizing and Optimizing LSTM Language Models",
               arXiv:1708.02182v1, 2017.

        .. [2] Y. Gal, Z. Ghahramani "A Theoretically Grounded Application of Dropout
               in Recurrent Neural Networks" arXiv:1512.05287v5, 2016. """
    if s0 is not None:
        assert isinstance(s0, np.ndarray) or (isinstance(s0, Tensor) and s0.constant is True)
    s = Tensor._op(GRUnit, X, Uz, Wz, bz, Ur, Wr, br, Uh, Wh, bh,
                   op_kwargs=dict(s0=s0, bp_lim=bp_lim, dropout=dropout),
                   constant=constant)
    s.creator._hidden_seq = s
    return s



import versioneer
from setuptools import find_packages, setup

DISTNAME = 'mygrad'
LICENSE = 'MIT'
AUTHOR = 'Ryan Soklaski'
AUTHOR_EMAIL = 'rsoklaski@gmail.com'
URL = 'https://github.com/rsokl/MyGrad'
CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Intended Audience :: Science/Research',
    'Intended Audience :: Education',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Topic :: Scientific/Engineering',
]

INSTALL_REQUIRES = ['numpy >= 1.12']
TESTS_REQUIRE = ['pytest >= 3.8', 'hypothesis >= 4.0', 'scipy']

DESCRIPTION = "A sleek auto-differentiation library that wraps numpy."
LONG_DESCRIPTION = """
mygrad is a simple, NumPy-centric autograd library. An autograd library enables 
you to automatically compute derivatives of mathematical functions. This library is 
designed to serve primarily as an education tool for learning about gradient-based 
machine learning; it is easy to install, has a readable and easily customizable code base, 
and provides a sleek interface that mimics NumPy. Furthermore, it leverages NumPy's 
vectorization to achieve good performance despite the library's simplicity.

This is not meant to be a competitor to libraries like PyTorch (which mygrad most 
closely resembles) or TensorFlow. Rather, it is meant to serve as a useful tool for 
students who are learning about training neural networks using back propagation.
"""


setup(name=DISTNAME,
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      license=LICENSE,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      classifiers=CLASSIFIERS,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      install_requires=INSTALL_REQUIRES,
      tests_require=TESTS_REQUIRE,
      extras_require={
          'rnn': ["numba>=0.34.0"],  # GRU and vanilla RNN require numba-acceleration
      },
      url=URL,
      download_url="https://github.com/rsokl/mygrad/tarball/"+versioneer.get_version(),
      python_requires='!=2.*,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*',
      packages=find_packages(exclude=["tests"]),
      )



import mygrad as mg
from mygrad.tensor_base import Tensor

import hypothesis.strategies as st
from numpy.testing import assert_allclose

from ...custom_strategies import valid_axes
from hypothesis import given
import hypothesis.extra.numpy as hnp

import numpy as np

dtype_strat_numpy = st.sampled_from((np.int8, np.int16, np.int32, np.int64,
                                     np.float16, np.float32, np.float64))


@given(a=hnp.arrays(shape=hnp.array_shapes(max_side=4, max_dims=5),
                    dtype=dtype_strat_numpy),
       data=st.data())
def test_argmin(a, data):
    axis = data.draw(valid_axes(ndim=a.ndim, single_axis_only=True), label="axis")
    tensor = Tensor(a)
    # tensor input
    assert_allclose(mg.argmin(tensor, axis=axis), np.argmin(a, axis=axis))
    
    # tensor method
    assert_allclose(tensor.argmin(axis=axis), a.argmin(axis=axis))
    
    # array input
    assert_allclose(mg.argmin(a, axis=axis), np.argmin(a, axis=axis))


@given(a=hnp.arrays(shape=hnp.array_shapes(max_side=4, max_dims=5),
                    dtype=dtype_strat_numpy),
       data=st.data())
def test_argmax(a, data):
    axis = data.draw(valid_axes(ndim=a.ndim, single_axis_only=True), label="axis")
    tensor = Tensor(a)
    
    # tensor input
    assert_allclose(mg.argmax(tensor, axis=axis), np.argmax(a, axis=axis))
    
    # tensor method
    assert_allclose(tensor.argmax(axis=axis), a.argmax(axis=axis))
    
    # array input
    assert_allclose(mg.argmax(a, axis=axis), np.argmax(a, axis=axis))



from tests.wrappers.uber import fwdprop_test_factory, backprop_test_factory

from mygrad import arcsin, arccos, arctan, arccsc, arcsec, arccot
import numpy as np


@fwdprop_test_factory(mygrad_func=arcsin, true_func=np.arcsin,
                      index_to_bnds={0: (-.9, .9)}, num_arrays=1)
def test_arcsin_fwd():
    pass


@backprop_test_factory(mygrad_func=arcsin, true_func=np.arcsin,
                       index_to_bnds={0: (-.9, .9)}, num_arrays=1,
                       as_decimal=False)
def test_arcsin_backward(data):
    pass


@fwdprop_test_factory(mygrad_func=arccos, true_func=np.arccos,
                      index_to_bnds={0: (-.9, .9)}, num_arrays=1)
def test_arccos_fwd():
    pass


@backprop_test_factory(mygrad_func=arccos, true_func=np.arccos,
                       index_to_bnds={0: (-.9, .9)}, num_arrays=1,
                       as_decimal=False)
def test_arccos_backward(data):
    pass


@fwdprop_test_factory(mygrad_func=arctan, true_func=np.arctan,
                      index_to_bnds={0: (.1, 10.)}, num_arrays=1)
def test_arctan_fwd():
    pass


@backprop_test_factory(mygrad_func=arctan, true_func=np.arctan,
                       index_to_bnds={0: (.1, 10.)}, num_arrays=1,
                       as_decimal=False)
def test_arctan_backward(data):
    pass


@fwdprop_test_factory(mygrad_func=arccsc, true_func=lambda x: np.arcsin(1 / x),
                      index_to_bnds={0: (-10.1, -1.1)}, num_arrays=1)
def test_arccsc_fwd():
    pass


@backprop_test_factory(mygrad_func=arccsc, true_func=lambda x: np.arcsin(1 / x),
                       index_to_bnds={0: (1.1, 100.)}, num_arrays=1,
                       as_decimal=False)
def test_arccsc_backward(data):
    pass


@fwdprop_test_factory(mygrad_func=arcsec, true_func=lambda x: np.arccos(1 / x),
                      index_to_bnds={0: (-10.1, -1.1)}, num_arrays=1)
def test_arcsec_fwd():
    pass


@backprop_test_factory(mygrad_func=arcsec, true_func=lambda x: np.arccos(1 / x),
                       index_to_bnds={0: (1.1, 100.)}, num_arrays=1,
                       as_decimal=False, atol=1e-5)
def test_arcsec_backward(data):
    pass


@fwdprop_test_factory(mygrad_func=arccot, true_func=lambda x: np.arctan(1 / x),
                      index_to_bnds={0: (0.1, 10)}, num_arrays=1)
def test_arccot_fwd():
    pass


@backprop_test_factory(mygrad_func=arccot, true_func=lambda x: np.arctan(1 / x),
                       index_to_bnds={0: (0.1, 10)}, num_arrays=1,
                       as_decimal=False, atol=1e-5)
def test_arccot_backward(data):
    pass



from mygrad import Tensor
import mygrad as mg
from mygrad.nnet.layers.batchnorm import batchnorm
import numpy as np
from numpy.testing import assert_allclose, assert_array_equal
import hypothesis.strategies as st
from hypothesis import given
import hypothesis.extra.numpy as hnp


def simple_batchnorm(x, gamma, beta, eps):
    axes = [i for i in range(x.ndim)]
    axes.pop(1)  # every axis except 1
    axes = tuple(axes)
    keepdims_shape = tuple(1 if n != 1 else d for n, d in enumerate(x.shape))

    mean = mg.mean(x, axis=axes, keepdims=True)
    var = mg.var(x, axis=axes, keepdims=True)
    norm = (x - mean) / mg.sqrt(var + eps)

    if gamma is not None:
        gamma = gamma.reshape(keepdims_shape)
        norm *= gamma

    if beta is not None:
        beta = beta.reshape(keepdims_shape)
        norm += beta
    return norm


@given(x=hnp.arrays(shape=hnp.array_shapes(min_dims=2, max_dims=4),
                    dtype=float,
                    elements=st.floats(-100, 100)),
       data=st.data())
def test_batchnorm(x, data):
    # optionally draw affine parameters
    gamma = data.draw(st.one_of(st.none(),
                                hnp.arrays(shape=x.shape[1:2], dtype=float, elements=st.floats(-10, 10))),
                      label="gamma")
    beta = data.draw(st.one_of(st.none(),
                               hnp.arrays(shape=x.shape[1:2], dtype=float, elements=st.floats(-10, 10))),
                     label="beta")
    x_orig = np.copy(x)

    gamma_orig = np.copy(gamma) if gamma is not None else None
    beta_orig = np.copy(beta) if beta is not None else None

    t1 = Tensor(x)
    t2 = Tensor(x)

    g1 = Tensor(gamma) if gamma is not None else None
    g2 = Tensor(gamma) if gamma is not None else None

    b1 = Tensor(beta) if beta is not None else None
    b2 = Tensor(beta) if beta is not None else None

    y1 = simple_batchnorm(t1, g1, b1, eps=1e-6)
    y2 = batchnorm(t2, gamma=g2, beta=b2, eps=1e-6)

    assert_allclose(actual=y2.data, desired=y1.data, atol=1e-6, rtol=1e-6)
    grad = data.draw(hnp.arrays(shape=y2.shape, dtype=t2.dtype, elements=st.floats(-10, 10)),
                     label='grad')
    grad_orig = np.copy(grad)
    
    y1.backward(grad)
    y2.backward(grad)

    assert_allclose(actual=t2.grad, desired=t1.grad, atol=1e-4, rtol=1e-4)

    if beta is not None:
        assert_allclose(actual=b2.grad, desired=b1.grad, atol=1e-4, rtol=1e-4)
    else:
        assert b2 is None

    if gamma is not None:
        assert_allclose(actual=g2.grad, desired=g1.grad, atol=1e-4, rtol=1e-4)
    else:
        assert g2 is None

    for n, (o, c) in enumerate(zip((x, gamma, beta, grad), (x_orig, gamma_orig, beta_orig, grad_orig))):
        if o is None or c is None:
            assert o is c, "('{x}', '{gamma}', '{beta}', '{grad}')[{n}]".format(x=x, gamma=gamma,
                                                                                  beta=beta, grad=grad, n=n)
        else:
            assert_array_equal(o, c, err_msg="('{x}', '{gamma}', '{beta}', '{grad}')[{n}]".format(x=x, gamma=gamma,
                                                                                                  beta=beta, grad=grad,
                                                                                                  n=n))

    if gamma is not None and beta is not None:
        assert not np.shares_memory(g2.grad, b2.grad)
    assert not np.shares_memory(grad, t2.grad)

    y2.null_gradients()
    assert t2.grad is None

    if gamma is not None:
        assert g2.grad is None

    if beta is not None:
        assert b2.grad is None




from tests.wrappers.uber import fwdprop_test_factory, backprop_test_factory
from mygrad import arcsinh, arccosh, arctanh, arccsch, arccoth
import numpy as np


def _is_non_zero(x):
    return np.all(np.abs(x.data) > 1e-8)


@fwdprop_test_factory(mygrad_func=arcsinh, num_arrays=1,  true_func=np.arcsinh)
def test_arcsinh_fwd():
    pass


@backprop_test_factory(mygrad_func=arcsinh, true_func=np.arcsinh, num_arrays=1, as_decimal=False)
def test_arcsinh_backward():
    pass


@fwdprop_test_factory(mygrad_func=arccosh, num_arrays=1,  true_func=np.arccosh,
                      index_to_bnds={0: (1.001, 10)})
def test_arccosh_fwd():
    pass


@backprop_test_factory(mygrad_func=arccosh, true_func=np.arccosh, num_arrays=1,
                       index_to_bnds={0: (1.001, 10)}, as_decimal=False)
def test_arccosh_backward():
    pass


@fwdprop_test_factory(mygrad_func=arctanh, true_func=np.arctanh,
                      num_arrays=1,  index_to_bnds={0: (-.5, .5)})
def test_arctanh_fwd():
    pass


@backprop_test_factory(mygrad_func=arctanh, true_func=np.arctanh, num_arrays=1,
                       index_to_bnds={0: (-0.5, 0.5)},
                       assumptions=_is_non_zero, as_decimal=False)
def test_arctanh_backward():
    pass


@fwdprop_test_factory(mygrad_func=arccsch, num_arrays=1,
                      true_func=lambda x: np.arcsinh(1 / x), index_to_bnds={0: (1, 10)})
def test_arccsch_fwd():
    pass


@backprop_test_factory(mygrad_func=arccsch, true_func=lambda x: np.arcsinh(1 / x),
                       num_arrays=1,  index_to_bnds={0: (1, 10)}, as_decimal=False)
def test_arccsch_backward():
    pass


@fwdprop_test_factory(mygrad_func=arccoth, num_arrays=1,
                      true_func=lambda x: np.arctanh(1 / x), index_to_bnds={0: (5, 10)})
def test_arccoth_fwd():
    pass


@backprop_test_factory(mygrad_func=arccoth, true_func=lambda x: np.arctanh(1 / x),
                       num_arrays=1,  index_to_bnds={0: (5, 10)}, as_decimal=False)
def test_arccoth_backward():
    pass


import numpy as np
from numpy.testing import assert_array_equal
from mygrad.tensor_creation.funcs import *
from mygrad import Tensor
from hypothesis import given
import hypothesis.strategies as st


def check_tensor_array(tensor, array, constant):
    assert isinstance(tensor, Tensor)
    assert_array_equal(tensor.data, array)
    assert tensor.dtype is array.dtype
    assert tensor.constant is constant


@given(constant=st.booleans(),
       dtype=st.sampled_from((np.int32, np.float64)))
def test_all_tensor_creation(constant, dtype):
    x = np.array([1, 2, 3])

    e = empty((3, 2), dtype=dtype, constant=constant)
    assert e.shape == (3, 2)
    assert e.constant is constant

    e = empty_like(e, dtype=dtype, constant=constant)
    assert e.shape == (3, 2)
    assert e.constant is constant

    check_tensor_array(eye(3, dtype=dtype, constant=constant), 
                       np.eye(3, dtype=dtype), constant)

    check_tensor_array(identity(3, dtype=dtype, constant=constant), 
                       np.identity(3, dtype=dtype), constant)

    check_tensor_array(ones((4, 5, 6), dtype=dtype, constant=constant), 
                       np.ones((4, 5, 6), dtype=dtype), constant)

    check_tensor_array(ones_like(x, dtype=dtype, constant=constant), 
                       np.ones_like(x, dtype=dtype), constant)

    check_tensor_array(zeros((4, 5, 6), dtype=dtype, constant=constant), 
                       np.zeros((4, 5, 6), dtype=dtype), constant)

    check_tensor_array(zeros_like(x, dtype=dtype, constant=constant), 
                       np.zeros_like(x, dtype=dtype), constant)
    
    check_tensor_array(full((4, 5, 6), 5., dtype=dtype, constant=constant),
                       np.full((4, 5, 6), 5., dtype=dtype), constant)

    check_tensor_array(full_like(x, 5., dtype=dtype, constant=constant),
                       np.full_like(x, 5., dtype=dtype), constant)

    check_tensor_array(arange(3, 7, dtype=dtype, constant=constant),
                       np.arange(3, 7, dtype=dtype), constant)

    check_tensor_array(linspace(3, 7, dtype=dtype, constant=constant),
                       np.linspace(3, 7, dtype=dtype), constant)

    check_tensor_array(logspace(3, 7, dtype=dtype, constant=constant),
                       np.logspace(3, 7, dtype=dtype), constant)

    check_tensor_array(geomspace(3, 7, dtype=dtype, constant=constant),
                       np.geomspace(3, 7, dtype=dtype), constant)
